"""
obspy.core.event.resourceid - ResourceIdentifier
================================================
This module defines the ResourceIdentifier class and associated code.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import re
import warnings
from contextlib import contextmanager
from copy import deepcopy
from uuid import uuid4
from weakref import WeakKeyDictionary, WeakValueDictionary

from obspy.core.util.decorator import deprecated


class _ResourceKey(object):
    """
    A private semi-singleton class used to refer id strings to objects.

    Only one instance should be created for each unique string/int, unless
    all reference are garbage collect, in which case a new instance will be
    created. Constructing instances through the get_resource_key class method
    is required to ensure this behavior.

    This class allows the python gc to handle cleanup of the
    ResourceIdentifier stored class state rather than manual reference
    counting as was implemented before.
    """
    # define a mapping from a resource string to a singleton instance
    _singleton_cache = WeakValueDictionary()

    def __deepcopy__(self, memodict={}):
        memodict[id(self)] = self
        return self

    def __copy__(self):
        return self

    @classmethod
    def get_resource_key(cls, unique_id):
        if unique_id not in _ResourceKey._singleton_cache:
            single = _ResourceKey()
            _ResourceKey._singleton_cache[unique_id] = single
        return _ResourceKey._singleton_cache[unique_id]


class _ResourceKeyDescriptor(object):
    """
    A private descriptor for initializing _Resource_Key instances.
    """

    def __init__(self, name, default=None):
        self.name = name + '__'
        self.default = default

    def __get__(self, instance, owner):
        return getattr(instance, self.name, self.default)

    def __set__(self, instance, value):
        # if an object was passed, use the id of the object for hash
        if value is not None:
            if not isinstance(value, (int, str, native_str)):
                value = id(value)
            setattr(instance, self.name, _ResourceKey.get_resource_key(value))


class ResourceIdentifier(object):
    r"""
    Unique identifier referring to a resource.

    In QuakeML many elements and types can have a unique id that other elements
    use to refer to it. This is called a ResourceIdentifier and it is used for
    the same purpose in the obspy.core.event classes. The id must be a string.

    In QuakeML it has to be of the following regex form::

        (smi|quakeml):[\w\d][\w\d\-\.\*\(\)_~']{2,}/[\w\d\-\.\*\(\)_~']
        [\w\d\-\.\*\(\)\+\?_~'=,;#/&]*

    e.g.

    * ``smi:sub.website.org/event/12345678``
    * ``quakeml:google.org/pick/unique_pick_id``

    smi stands for "seismological meta-information".

    :type id: str, optional
    :param id: A string to uniquely identify a resource. If no resource_id
        is given, uuid.uuid4() will be used to create one which assures
        uniqueness within the current Python run. If no fixed id is provided,
        the ID will be built from prefix and a random uuid hash. The random
        hash can be regenerated by the referred object automatically if it
        gets changed.
    :type prefix: str, optional
    :param prefix: An optional identifier that will be put in front of any
        automatically created resource id. The prefix will only have an effect
        if `id` is not specified (for a fixed ID string). Makes automatically
        generated resource ids more reasonable. By default "smi:local" is used
        which ensures a QuakeML compliant resource identifier.
    :type referred_object: Python object, optional
    :param referred_object: The object (resource) to which this instance
        refers. All instances created with the same resource_id will be able
        to access the object as long as at least one instance actually has a
        reference to it. Additionally, ResourceIdentifier instances that have
        the same id but refer to different objects will still return the
        correct referred object, provided it doesn't get garbage collected.
        If the referred object no longer exists the last object registered
        with the same id will be returned, or None if no such object exists.
    :type parent: hashable
    :param parent:
        The parent to use as a namespace for the resource identifier. This
        allows potentially interconnected resource identifiers to be sensibly
        grouped together, ensuring that objects belonging to the same parent
        are returned by resource ids of the parent. Used primarily to ensure
        resource_ids in an event refer to other objects in the same event.
        All resource ids must be unique within the parent namespace.

    .. rubric:: General Usage

    >>> ResourceIdentifier('2012-04-11--385392')
    ResourceIdentifier(id="2012-04-11--385392")
    >>> # If 'id' is not specified it will be generated automatically.
    >>> ResourceIdentifier()  # doctest: +ELLIPSIS
    ResourceIdentifier(id="smi:local/...")
    >>> # Supplying a prefix will simply prefix the automatically generated ID
    >>> ResourceIdentifier(prefix='event')  # doctest: +ELLIPSIS
    ResourceIdentifier(id="event/...")

    ResourceIdentifiers can, and oftentimes should, carry a reference to the
    object they refer to. This is a weak reference which means that if the
    object get deleted or runs out of scope, e.g. gets garbage collected, the
    reference will cease to exist.

    >>> from obspy.core.event import Event
    >>> event = Event()
    >>> import sys
    >>> ref_count = sys.getrefcount(event)
    >>> res_id = ResourceIdentifier(referred_object=event)
    >>> # The reference does not changed the reference count of the object.
    >>> print(ref_count == sys.getrefcount(event))
    True
    >>> # It actually is the same object.
    >>> print(event is res_id.get_referred_object())
    True
    >>> # Deleting it, or letting the garbage collector handle the object will
    >>> # invalidate the reference.
    >>> del event
    >>> print(res_id.get_referred_object())
    None

    The most powerful ability (and reason why one would want to use a resource
    identifier class in the first place) is that once a ResourceIdentifier with
    an attached referred object has been created, any other ResourceIdentifier
    instances with the same ID can retrieve that object. This works
    across all ResourceIdentifiers that have been instantiated within one
    Python run.
    This enables, e.g. the resource references between the different QuakeML
    elements to work in a rather natural way.

    >>> event_object = Event()
    >>> obj_id = id(event_object)
    >>> res_id = "obspy.org/event/test"
    >>> ref_a = ResourceIdentifier(res_id)
    >>> # The object is refers to cannot be found yet. Because no instance that
    >>> # an attached object has been created so far.
    >>> print(ref_a.get_referred_object())
    None
    >>> # This instance has an attached object.
    >>> ref_b = ResourceIdentifier(res_id, referred_object=event_object)
    >>> ref_c = ResourceIdentifier(res_id)
    >>> # All ResourceIdentifiers will refer to the same object.
    >>> assert ref_a.get_referred_object() is event_object
    >>> assert ref_b.get_referred_object() is event_object
    >>> assert ref_c.get_referred_object() is event_object

    Resource identifiers are bound to an object once the get_referred_object
    method has been called. The results is that get_referred_object will
    always return the same object it did on the first call as long as the
    object still exists. If the bound object gets garage collected a warning
    will be issued and another object with the same resource_id will be
    returned (if one exists). If no other object is associated with the same
    resource_id, an additional warning will be issued and None returned.

    >>> from obspy import UTCDateTime
    >>> res_id = 'obspy.org/tests/test_resource_doc_example'
    >>> obj_a = UTCDateTime(10)
    >>> obj_b = UTCDateTime(10)
    >>> ref_a = ResourceIdentifier(res_id, referred_object=obj_a)
    >>> ref_b = ResourceIdentifier(res_id, referred_object=obj_b)
    >>> assert ref_a.get_referred_object() == ref_b.get_referred_object()
    >>> assert ref_a.get_referred_object() is not ref_b.get_referred_object()
    >>> assert ref_a.get_referred_object() is obj_a
    >>> assert ref_b.get_referred_object() is obj_b
    >>> del obj_b  # obj_b gets garbage collected
    >>> assert ref_b.get_referred_object() is obj_a
    >>> del obj_a  # now no object with res_id exists
    >>> assert ref_b.get_referred_object() is None

    The id can be converted to a valid QuakeML ResourceIdentifier by calling
    the convert_id_to_quakeml_uri() method. The resulting id will be of the
    form::
        smi:authority_id/prefix/id

    >>> res_id = ResourceIdentifier(prefix='origin')
    >>> res_id.convert_id_to_quakeml_uri(authority_id="obspy.org")
    >>> res_id  # doctest: +ELLIPSIS
    ResourceIdentifier(id="smi:obspy.org/origin/...")
    >>> res_id = ResourceIdentifier(id='foo')
    >>> res_id.convert_id_to_quakeml_uri()
    >>> res_id
    ResourceIdentifier(id="smi:local/foo")
    >>> # A good way to create a QuakeML compatibly ResourceIdentifier from
    >>> # scratch is
    >>> res_id = ResourceIdentifier(prefix='pick')
    >>> res_id.convert_id_to_quakeml_uri(authority_id='obspy.org')
    >>> res_id  # doctest: +ELLIPSIS
    ResourceIdentifier(id="smi:obspy.org/pick/...")
    >>> # If the given ID is already a valid QuakeML
    >>> # ResourceIdentifier, nothing will happen.
    >>> res_id = ResourceIdentifier('smi:test.org/subdir/id')
    >>> res_id
    ResourceIdentifier(id="smi:test.org/subdir/id")
    >>> res_id.convert_id_to_quakeml_uri()
    >>> res_id
    ResourceIdentifier(id="smi:test.org/subdir/id")

    ResourceIdentifiers are considered identical if the IDs are
    the same.

    >>> # Create two different resource identifiers.
    >>> res_id_1 = ResourceIdentifier()
    >>> res_id_2 = ResourceIdentifier()
    >>> res_id_3 = ResourceIdentifier(id=res_id_2.id)
    >>> assert res_id_1 != res_id_2
    >>> assert res_id_2 == res_id_3

    ResourceIdentifier instances can be used as dictionary keys.

    >>> dictionary = {}
    >>> res_id = ResourceIdentifier(id="foo")
    >>> dictionary[res_id] = "bar1"
    >>> # The same ID can still be used as a key.
    >>> dictionary["foo"] = "bar2"
    >>> items = sorted(dictionary.items(), key=lambda kv: kv[1])
    >>> for k, v in items:  # doctest: +ELLIPSIS
    ...     print(repr(k), v)
    ResourceIdentifier(id="foo") bar1
    ...'foo' bar2

    Because ResourceIdentifier instances are hashed based on their id
    attribute, you should never change it once it has been set. Create a new
    ResourceIdentifier object instead.
    """
    # Class (not instance) attributes that keeps track of all resource
    # identifier instances throughout one Python run. Will only store weak
    # references and therefore does not interfere with the garbage collection.
    # DO NOT CHANGE THIS FROM OUTSIDE THE CLASS. Use the _debug_class_state
    # method for temporarily altering these for testing or debugging.

    # A dict for keeping track of parent/object specific references.
    # weak_key, weak_key, weak_value
    # {parent_key: {resource_key: object_id}}
    _parent_id_tree = WeakKeyDictionary()

    # Dict of list to keep track of the order in which resource_ids are
    # assigned to object_ids.
    # {resource_id: [object_id, object_id, ...]}
    _id_order = WeakKeyDictionary()

    # A weak value dict mapping object ids (not resource ids) to objects.
    # {object_id: object}
    _id_object_map = WeakValueDictionary()

    # A get objects hook to allow plugging in custom logic for finding objects
    # with by resource_id if they are not found via normal means.
    _get_object_hook = []

    # Set default _ResourceKey attributes and object_id.
    _parent_key = _ResourceKeyDescriptor('_parent_key')
    _resource_key = _ResourceKeyDescriptor('_resource_key')
    _object_id = None

    def __init__(self, id=None, prefix="smi:local", referred_object=None,
                 parent=None):
        # Create a resource id if None is given and possibly use a prefix.
        if id is None:
            self.fixed = False
            self._prefix = prefix
            self._uuid = str(uuid4())
        elif isinstance(id, ResourceIdentifier):
            self.__dict__.update(id.__dict__)
            return
        else:
            self.fixed = True
            self.id = id
        # get resource and parent _ResourceKey singletons
        self._resource_key = self.id
        self._parent_key = parent
        # set referred object if one was provided.
        if referred_object is not None:
            self.set_referred_object(referred_object)

    def get_referred_object(self):
        """
        Returns the object associated with the resource identifier.

        This works as long as at least one ResourceIdentifier with the same
        ID as this instance has an associated object. If not, this method will
        return None.
        """
        try:
            out = ResourceIdentifier._id_object_map[self._object_key]
        except KeyError:
            out = self._get_similar_referred_object()
        if out is not None:
            return out
        else:  # If no object was found iterate get_object hooks.
            for hook in self._get_object_hook:
                out = hook(self.id)
                if out is not None:
                    self.set_referred_object(out)
                    return out

    def _get_similar_referred_object(self):
        """
        Find an object with the same resource id.

        If the resource_identifier instance is not bound to a specific object,
        or the bound object has been garbage collected, try to find another
        object that was assigned the same resource_id.

        The parent_id_tree will be scanned first to see if any object_ids
        have been assigned to the same resource_id in the parent scope. If
        not simply use the last object assigned the resource_id. If no object
        is found return None.
        """
        # Warn if resource_id was bound but its object got gc'ed.
        if self._object_id is not None:
            msg = ("The object with identity of: %d  and id of %s no longer "
                   "exists, trying to find an object with the same id"
                   ) % (self._object_id, self.id)
            warnings.warn(msg, UserWarning)
        # try to get the object_id from the parent_id_tree
        obj = self._get_object_from_parent_scope()
        if obj is None:
            # else try to find the last object assigned the same resource_id
            obj = self._get_newest_object_with_same_resource_id()
        # if an object was found bind it to resource_id and return it
        if obj is not None:
            self.set_referred_object(obj, warn=False)
        return obj

    def _get_newest_object_with_same_resource_id(self):
        """
        Return the newest object which has been bound to the same resource_id.
        If No such object exists return None.
        """
        try:
            rid_list = ResourceIdentifier._id_order[self._resource_key]
        except KeyError:  # no existing object is assigned to this resource_id
            return None
        while len(rid_list):
            try:
                obj_key = ResourceIdentifier._id_order[self._resource_key][-1]
            except KeyError:
                return None
            else:
                try:
                    obj = ResourceIdentifier._id_object_map[obj_key]
                except KeyError:  # Object got gc'ed or was not defined.
                    rid_list.pop()  # Remove last element in list.
                    continue
                # Ensure the current resource_id matches the old one, there
                # can, in rare cases, be a mismatch due to issue #2278.
                if obj_key[-1] != self.id:
                    rid_list.pop()
                    continue
                return obj

    def _get_object_from_parent_scope(self):
        """
        Find an object in the same parent scope with the same resource_id. If
        No such object is found return None.
        """
        id_tree = ResourceIdentifier._parent_id_tree
        try:
            obj_key = id_tree[self._parent_key][self._resource_key]
        except (KeyError, TypeError):
            pass
        else:
            try:
                obj = ResourceIdentifier._id_object_map[obj_key]
            except KeyError:
                pass
            else:
                if obj_key[-1] == self.id:
                    return obj

    def set_referred_object(self, referred_object, warn=True, parent=None):
        """
        Bind an object to the ResourceIdentifier instance.

        Also allows the object to be bound to a parent scope. Will emit a
        warning if referred_object is not equal to the last object referred
        to by the same resource_id, or, if the resource_id does not refer to
        an object that currently exists, the last object assigned to the same
        resource id code.

        :param referred_object: The object to which the resource id refers.
        :type referred_object: object
        :param warn:
            If True, issue a warning if the referred_object is not equal to
            the last object assigned the same resource_id.
        :type warn: bool
        :param parent:
            An object or int (id) to which the resource_id should be scoped.
            Used, for example, to ensure the all resource ids belonging to an
            event object are event-scoped.
        :type parent: object, int
        """

        # Get the last object bound to this instance of ResourceIdentifier or
        # if there is None, get the last referred_object assigned the same
        # resource_id code.
        id_order = ResourceIdentifier._id_order
        old = ResourceIdentifier._id_object_map.get(self._object_key, None)
        if old is None:  # Look for last object with same resource id.
            try:
                old_obj_id_key = id_order[self._resource_key][-1]
                old = ResourceIdentifier._id_object_map[old_obj_id_key]
            except (KeyError, IndexError):
                pass
        if warn and old is not None and old != referred_object:
            msg = ('Warning, binding object to resource ID %s which '
                   'is not equal to the last object bound to this '
                   'resource_id') % self.id
            warnings.warn(msg, UserWarning)
        # Set the object id to the new object, and update parent scoping tree.
        self._object_id = id(referred_object)
        if parent is not None or self._parent_key is not None:
            if parent is not None:
                self._parent_key = parent
            id_tree = ResourceIdentifier._parent_id_tree
            if self._parent_key not in id_tree:
                id_tree[self._parent_key] = WeakKeyDictionary()
            id_tree[self._parent_key][self._resource_key] = self._object_key
        # Set the new id in id map and append referred_object to id_order.
        ResourceIdentifier._id_object_map[self._object_key] = referred_object
        if self._resource_key not in id_order:
            id_order[self._resource_key] = []
        id_order[self._resource_key].append(self._object_key)

    @deprecated()
    def convert_id_to_quakeml_uri(self, authority_id="local"):
        """
        Converts the current ID to a valid QuakeML URI.

        This method is deprecated, use :meth:`get_quakeml_id` instead.

        Only an invalid QuakeML ResourceIdentifier string it will be converted
        to a valid one.  Otherwise nothing will happen but after calling this
        method the user can be sure that the ID is a valid QuakeML URI.

        The resulting ID will be of the form
            smi:authority_id/prefix/resource_id

        :type authority_id: str, optional
        :param authority_id: The base url of the resulting string. Defaults to
            ``"local"``.
        """
        self.id = self.get_quakeml_uri_str(authority_id=authority_id)

    def get_quakeml_id(self, authority_id="local"):
        """
        Returns a resource id with a valid QuakeML URI.

        Only an invalid QuakeML ResourceIdentifier string it will be converted
        to a valid one.  Otherwise the returned resource id will be identical
        to the original.

        The new resource id will have the same referred object as the
        original.

        The resulting ID will be of the form
            smi:authority_id/prefix/resource_id

        :type authority_id: str, optional
        :param authority_id: The base url of the resulting string. Defaults to
            ``"local"``.
        :return: A new ResourceIdentifier instance with a valid quakeml uri.
        """
        new_id = self.get_quakeml_uri_str(authority_id=authority_id)
        rid = ResourceIdentifier(new_id)
        referred_obj = self.get_referred_object()
        if referred_obj is not None:
            rid.set_referred_object(referred_obj, warn=False,
                                    parent=self._parent_key)
        return rid

    @deprecated()
    def get_quakeml_uri(self, authority_id="local"):
        """
        This method is deprecated, use :meth:`get_quakeml_uri_str` instead.
        """
        return self.get_quakeml_uri_str(authority_id=authority_id)

    def get_quakeml_uri_str(self, authority_id="local"):
        """
        Returns an id with a valid QuakeML URI.

        If no valid QuakeML is possible a ValueError is raised.

        :type authority_id: str, optional
        :param authority_id: The base url of the resulting string. Defaults to
            ``"local"``.
        :return: A new ResourceIdentifier instance with a valid quakeml uri.

        >>> res_id = ResourceIdentifier("some_id")
        >>> print(res_id.get_quakeml_uri_str())
        smi:local/some_id
        >>> # Did not change the actual resource id.
        >>> print(res_id.id)
        some_id
        """
        id = self.id
        if str(id).strip() == "":
            id = str(uuid4())

        regex = r"^(smi|quakeml):[\w\d][\w\d\-\.\*\(\)_~']{2,}/[\w\d\-\." + \
                r"\*\(\)_~'][\w\d\-\.\*\(\)\+\?_~'=,;#/&]*$"
        result = re.match(regex, str(id))
        if result is not None:
            return id
        id = 'smi:%s/%s' % (authority_id, str(id))
        # Check once again just to be sure no weird symbols are stored in the
        # ID.
        result = re.match(regex, id)
        if result is None:
            msg = (
                "The id '%s' is not a valid QuakeML resource "
                "identifier. ObsPy tried modifying it to '%s' but it is still "
                "not valid. Please make sure all resource ids are either "
                "valid or can be made valid by prefixing them with "
                "'smi:<authority_id>/'. Valid ids are specified in the "
                "QuakeML manual section 3.1 and in particular exclude colons "
                "for the final part." % (self.id, id))
            raise ValueError(msg)
        return id

    def copy(self):
        """
        Returns a copy of the ResourceIdentifier.
        >>> res_id = ResourceIdentifier()
        >>> res_id_2 = res_id.copy()
        >>> print(res_id is res_id_2)
        False
        >>> print(res_id == res_id_2)
        True
        """
        return deepcopy(self)

    def __deepcopy__(self, memodict={}):
        new = ResourceIdentifier.__new__(ResourceIdentifier)
        new.__dict__.update(self.__dict__)
        # clear object_id upon copying resource_ids
        new._parent_key = None
        memodict[id(self)] = new
        return new

    def __setstate__(self, state):
        """
        Make sure the resource_key follows the singleton pattern.
        """
        self.__dict__ = state
        self._parent_key = None
        self._resource_key = _ResourceKey.get_resource_key(self.id)

    @property
    def _object_key(self):
        """
        The value used to identify objects bound to resource_ids.

        Uses a hash of both the object id and resource id, see #2278.
        """
        return self._object_id, self.id

    @property
    def id(self):
        """
        Unique identifier of the current instance.
        """
        if self.fixed:
            return self.__dict__.get("id")
        else:
            id = self.prefix
            if not id.endswith("/"):
                id += "/"
            id += self.uuid
            return id

    @id.deleter
    def id(self):
        msg = "The resource id cannot be deleted."
        raise Exception(msg)

    @id.setter
    def id(self, value):
        self.fixed = True
        # XXX: no idea why I had to add bytes for PY2 here
        if not isinstance(value, (str, bytes)):
            msg = "attribute id needs to be a string."
            raise TypeError(msg)
        if '_resource_key__' in self.__dict__:
            msg = ('overwritting the id attribute of a ResourceIdentifier'
                   'object is very dangerous and will raise an exception in '
                   'a future version of obspy')
            warnings.warn(msg, UserWarning)
        self.__dict__["id"] = value

    @property
    def prefix(self):
        return self._prefix

    @prefix.deleter
    def prefix(self):
        self._prefix = ""

    @prefix.setter
    def prefix(self, value):
        if not isinstance(value, (str, native_str)):
            msg = "prefix id needs to be a string."
            raise TypeError(msg)
        self._prefix = value

    @property
    def uuid(self):
        return self._uuid

    @uuid.deleter
    def uuid(self):
        """
        Deleting is uuid hash is forbidden and will not work.
        """
        msg = "The uuid cannot be deleted."
        raise Exception(msg)

    @uuid.setter
    def uuid(self, value):  # @UnusedVariable
        """
        Setting is uuid hash is forbidden and will not work.
        """
        msg = "The uuid cannot be set manually."
        raise Exception(msg)

    @property
    def resource_id(self):
        return self.id

    @resource_id.deleter
    def resource_id(self):
        del self.id

    @resource_id.setter
    def resource_id(self, value):
        self.id = value

    def __str__(self):
        return self.id

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))

    def __repr__(self):
        return 'ResourceIdentifier(id="%s")' % self.id

    def __eq__(self, other):
        if self.id == other:
            return True
        if not isinstance(other, ResourceIdentifier):
            return False
        if self.id == other.id:
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        """
        Uses the same hash as the resource id. This means that class instances
        can be used in dictionaries and other hashed types.
        Both the object and it's id can still be independently used as
        dictionary keys.
        """
        # "Salt" the hash with a string so the hash of the object and a
        # string identical to the id can both be used as individual
        # dictionary keys.
        return hash("RESOURCE_ID") + self.id.__hash__()

    # Like the weakref.ref class, calling a resource id instance can
    # return the referred object, if in scope, else None.
    __call__ = get_referred_object

    @deprecated()
    def regenerate_uuid(self):
        """
        Regenerates the uuid part of the ID. Does nothing for resource
        identifiers with a user-set, fixed id.
        """
        self._uuid = str(uuid4())

    @classmethod
    def register_get_object_hook(cls, get_object_callable):
        """
        Register a callable to return referred object if normal means fail.

        This is useful, for example, to allow the resource_identifier to
        create objects from entries in a database that do not yet have an
        in-memory representation.

        :param get_object_callable:
            Any callable that takes a resource id string and returns an
            object.
        :type get_object_callable: callable
        """
        cls._get_object_hook.append(get_object_callable)

    @classmethod
    def remove_get_object_hook(cls, get_object_callable):
        """
        Remove a callable from the registered get_object hooks.

        :param get_object_callable:
            The callable to remove from the get_object hook registry.
        :type get_object_callable: callable, optional
        """
        if get_object_callable is not None:
            try:
                cls._get_object_hook.remove(get_object_callable)
            except ValueError:  # Pass if get_object_callable is not in list.
                pass
        else:
            cls._get_object_hook.clear()

    @classmethod
    def _bind_class_state(cls, state_dict):
        """
        Bind the state contained in state_dict to ResourceIdentifier class.
        """
        cls._parent_id_tree = state_dict['parent_id_tree']
        cls._id_order = state_dict['id_order']
        cls._id_object_map = state_dict['id_object_map']
        cls._get_referred_object_hook = state_dict['get_object_hook']

    @classmethod
    @contextmanager
    def _debug_class_state(cls):
        """
        Context manager for debugging the class level state for Resource Ids.

        Replaces the current resource_id and unbound mappings returning
        a dictionary with the new mappings as values and "rdict", and
        "unbound" as keys. This function restores original mappings upon exit.
        """
        # get current class state
        old_state = dict(
            parent_id_tree=cls._parent_id_tree,
            id_order=cls._id_order,
            id_object_map=cls._id_object_map,
            get_object_hook=cls._get_object_hook,
        )
        # init new class state
        new_state = dict(
            parent_id_tree=WeakKeyDictionary(),
            id_order=WeakKeyDictionary(),
            id_object_map=WeakValueDictionary(),
            get_object_hook=[],
        )
        # bind new state and return dict
        cls._bind_class_state(new_state)
        yield new_state
        # reset prior state
        cls._bind_class_state(old_state)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.io.css.core test suite.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import gzip
import os
import unittest

import numpy as np

from obspy import read
from obspy.core import Stream, Trace, UTCDateTime
from obspy.core.util import NamedTemporaryFile
from obspy.io.css.core import (_is_css, _read_css, _is_nnsa_kb_core,
                               _read_nnsa_kb_core)


class CoreTestCase(unittest.TestCase):
    """
    Test cases for css core interface
    """
    def setUp(self):
        # directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data')
        self.filename_css = os.path.join(self.path, 'test_css.wfdisc')
        self.filename_nnsa = os.path.join(self.path, 'test_nnsa.wfdisc')
        # set up stream for validation
        header = {}
        header['station'] = 'TEST'
        header['starttime'] = UTCDateTime(1296474900.0)
        header['sampling_rate'] = 80.0
        header['calib'] = 1.0
        header['calper'] = 1.0
        header['_format'] = 'CSS'
        filename = os.path.join(self.path, '201101311155.10.ascii.gz')
        with gzip.open(filename, 'rb') as fp:
            data = np.loadtxt(fp, dtype=np.int_)
        # traces in the test files are sorted ZEN
        st = Stream()
        for x, cha in zip(data.reshape((3, 4800)), ('HHZ', 'HHE', 'HHN')):
            # big-endian copy
            tr = Trace(x, header.copy())
            tr.stats.station += 'be'
            tr.stats.channel = cha
            st += tr
            # little-endian copy
            tr = Trace(x, header.copy())
            tr.stats.station += 'le'
            tr.stats.channel = cha
            st += tr
        self.st_result_css = st.copy()
        for tr in st:
            tr.stats['_format'] = "NNSA_KB_CORE"
        self.st_result_nnsa = st

    def test_is_css(self):
        """
        Read files via obspy.core.stream.read function.
        """
        # 1
        assert(_is_css(self.filename_css))
        # check that empty files are not recognized as CSS
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            fh = open(tempfile, "wb")
            fh.close()
            assert(not _is_css(tempfile))

    def test_is_nnsa_kb_core(self):
        """
        Read files via obspy.core.stream.read function.
        """
        # 1
        assert(_is_nnsa_kb_core(self.filename_nnsa))
        # check that empty files are not recognized as NNSA_KB_CORE
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            fh = open(tempfile, "wb")
            fh.close()
            assert(not _is_nnsa_kb_core(tempfile))

    def test_css_read_via_obspy(self):
        """
        Read files via obspy.core.stream.read function.
        """
        # 1
        st = read(self.filename_css)
        self.assertEqual(st, self.st_result_css)

    def test_css_read_via_module(self):
        """
        Read files via obspy.io.css.core._read_css function.
        """
        # 1
        st = _read_css(self.filename_css)
        # _format entry is not present when using low-level function
        for tr in self.st_result_css:
            tr.stats.pop('_format')
        self.assertEqual(st, self.st_result_css)

    def test_nnsa_kb_core_read_via_obspy(self):
        """
        Read files via obspy.core.stream.read function.
        """
        # 1
        st = read(self.filename_nnsa)
        self.assertEqual(st, self.st_result_nnsa)

    def test_nnsa_kb_core_read_via_module(self):
        """
        Read files via obspy.io.css.core._read_nnsa_kb_core function.
        """
        # 1
        st = _read_nnsa_kb_core(self.filename_nnsa)
        # _format entry is not present when using low-level function
        for tr in self.st_result_nnsa:
            tr.stats.pop('_format')
        self.assertEqual(st, self.st_result_nnsa)


def suite():
    return unittest.makeSuite(CoreTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
Client for a database created by obspy.db.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import os

from sqlalchemy import and_, create_engine, func, or_
from sqlalchemy.orm import sessionmaker

from obspy.core.preview import merge_previews
from obspy.core.stream import Stream
from obspy.core.utcdatetime import UTCDateTime
from obspy.db.db import Base, WaveformChannel, WaveformFile, WaveformPath


class Client(object):
    """
    Client for a database created by obspy.db.
    """
    def __init__(self, url=None, session=None, debug=False):
        """
        Initializes the client.

        :type url: str, optional
        :param url: A string that indicates database dialect and connection
            arguments. See
            http://docs.sqlalchemy.org/en/latest/core/engines.html for more
            information about database dialects and urls.
        :type session: :class:`sqlalchemy.orm.session.Session`, optional
        :param session: An existing database session object.
        :type debug: bool, optional
        :param debug: Enables verbose output.
        """
        if url:
            self.engine = create_engine(url, encoding=native_str('utf-8'),
                                        convert_unicode=True)
            Base.metadata.create_all(self.engine,  # @UndefinedVariable
                                     checkfirst=True)
            # enable verbosity after table creations
            self.engine.echo = debug
            self.session = sessionmaker(bind=self.engine)
        else:
            self.session = session

    def get_network_ids(self):
        """
        Fetches all possible network id's.
        """
        session = self.session()
        query = session.query(WaveformChannel.network)
        query = query.group_by(WaveformChannel.network)
        results = query.all()
        session.close()
        return [r[0] for r in results if len(r) == 1]

    def get_station_ids(self, network=None):
        """
        Fetches all possible station id's.

        :type network: str, optional
        :param network: Filter result by given network id if given. Defaults
            to ``None``.
        """
        session = self.session()
        query = session.query(WaveformChannel.station)
        if network:
            query = query.filter(WaveformChannel.network == network)
        query = query.group_by(WaveformChannel.station)
        results = query.all()
        session.close()
        return [r[0] for r in results if len(r) == 1]

    def get_location_ids(self, network=None, station=None):
        """
        Fetches all possible location id's.

        :type network: str, optional
        :param network: Filter result by given network id if given. Defaults
            to ``None``.
        :type station: str, optional
        :param station: Filter result by given station id if given. Defaults
            to ``None``.
        """
        session = self.session()
        query = session.query(WaveformChannel.location)
        if network:
            query = query.filter(WaveformChannel.network == network)
        if station:
            query = query.filter(WaveformChannel.station == station)
        query = query.group_by(WaveformChannel.location)
        results = query.all()
        session.close()
        return [r[0] for r in results if len(r) == 1]

    def get_channel_ids(self, network=None, station=None, location=None):
        """
        Fetches all possible channel id's.

        :type network: str, optional
        :param network: Filter result by given network id if given. Defaults
            to ``None``.
        :type station: str, optional
        :param station: Filter result by given station id if given. Defaults
            to ``None``.
        :type location: str, optional
        :param location: Filter result by given location id if given. Defaults
            to ``None``.
        """
        session = self.session()
        query = session.query(WaveformChannel.channel)
        if network:
            query = query.filter(WaveformChannel.network == network)
        if station:
            query = query.filter(WaveformChannel.station == station)
        if location:
            query = query.filter(WaveformChannel.location == location)
        query = query.group_by(WaveformChannel.channel)
        results = query.all()
        session.close()
        return [r[0] for r in results if len(r) == 1]

    def get_endtimes(self, network=None, station=None, location=None,
                     channel=None):
        """
        Generates a list of last end times for each channel.
        """
        # build up query
        session = self.session()
        query = session.query(
            WaveformChannel.network, WaveformChannel.station,
            WaveformChannel.location, WaveformChannel.channel,
            func.max(WaveformChannel.endtime).label('latency')
        )
        query = query.group_by(
            WaveformChannel.network, WaveformChannel.station,
            WaveformChannel.location, WaveformChannel.channel
        )
        # process arguments
        kwargs = {'network': network, 'station': station,
                  'location': location, 'channel': channel}
        for key, value in kwargs.items():
            if value is None:
                continue
            col = getattr(WaveformChannel, key)
            if '*' in value or '?' in value:
                value = value.replace('?', '_')
                value = value.replace('*', '%')
                query = query.filter(col.like(value))
            else:
                query = query.filter(col == value)
        results = query.all()
        session.close()
        adict = {}
        for result in results:
            key = '%s.%s.%s.%s' % (result[0], result[1], result[2], result[3])
            adict[key] = UTCDateTime(result[4])
        return adict

    def get_waveform_path(self, network=None, station=None, location=None,
                          channel=None, starttime=None, endtime=None):
        """
        Generates a list of available waveform files.
        """
        # build up query
        session = self.session()
        query = session.query(WaveformPath.path,
                              WaveformFile.file,
                              WaveformChannel.network,
                              WaveformChannel.station,
                              WaveformChannel.location,
                              WaveformChannel.channel)
        query = query.filter(WaveformPath.id == WaveformFile.path_id)
        query = query.filter(WaveformFile.id == WaveformChannel.file_id)
        # process arguments
        kwargs = {'network': network, 'station': station,
                  'location': location, 'channel': channel}
        for key, value in kwargs.items():
            if value is None:
                continue
            col = getattr(WaveformChannel, key)
            if '*' in value or '?' in value:
                value = value.replace('?', '_')
                value = value.replace('*', '%')
                query = query.filter(col.like(value))
            else:
                query = query.filter(col == value)
        # start and end time
        try:
            starttime = UTCDateTime(starttime)
        except Exception:
            starttime = UTCDateTime() - 60 * 20
        finally:
            query = query.filter(WaveformChannel.endtime > starttime.datetime)
        try:
            endtime = UTCDateTime(endtime)
        except Exception:
            # 10 minutes
            endtime = UTCDateTime()
        finally:
            query = query.filter(WaveformChannel.starttime < endtime.datetime)
        results = query.all()
        session.close()
        # execute query
        file_dict = {}
        for result in results:
            fname = os.path.join(result[0], result[1])
            key = '%s.%s.%s.%s' % (result[2], result[3], result[4], result[5])
            file_dict.setdefault(key, []).append(fname)
        return file_dict

    def get_preview(self, trace_ids=[], starttime=None, endtime=None,
                    network=None, station=None, location=None, channel=None,
                    pad=False):
        """
        Returns the preview trace.
        """
        # build up query
        session = self.session()
        query = session.query(WaveformChannel)
        # start and end time
        try:
            starttime = UTCDateTime(starttime)
        except Exception:
            starttime = UTCDateTime() - 60 * 20
        finally:
            query = query.filter(WaveformChannel.endtime > starttime.datetime)
        try:
            endtime = UTCDateTime(endtime)
        except Exception:
            # 10 minutes
            endtime = UTCDateTime()
        finally:
            query = query.filter(WaveformChannel.starttime < endtime.datetime)
        # process arguments
        if trace_ids:
            # filter over trace id list
            trace_filter = or_()
            for trace_id in trace_ids:
                temp = trace_id.split('.')
                if len(temp) != 4:
                    continue
                trace_filter.append(and_(
                    WaveformChannel.network == temp[0],
                    WaveformChannel.station == temp[1],
                    WaveformChannel.location == temp[2],
                    WaveformChannel.channel == temp[3]))
            if trace_filter.clauses:
                query = query.filter(trace_filter)
        else:
            # filter over network/station/location/channel id
            kwargs = {'network': network, 'station': station,
                      'location': location, 'channel': channel}
            for key, value in kwargs.items():
                if value is None:
                    continue
                col = getattr(WaveformChannel, key)
                if '*' in value or '?' in value:
                    value = value.replace('?', '_')
                    value = value.replace('*', '%')
                    query = query.filter(col.like(value))
                else:
                    query = query.filter(col == value)
        # execute query
        results = query.all()
        session.close()
        # create Stream
        st = Stream()
        for result in results:
            preview = result.get_preview()
            st.append(preview)
        # merge and trim
        st = merge_previews(st)
        st.trim(starttime, endtime, pad=pad)
        return st



# -*- coding: utf-8 -*-
"""
SQLAlchemy ORM definitions (database layout) for obspy.db.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import pickle

from sqlalchemy import (Boolean, Column, DateTime, Float, ForeignKey, Integer,
                        PickleType, String)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relation
from sqlalchemy.schema import UniqueConstraint
import numpy as np

from obspy import Trace, UTCDateTime


Base = declarative_base()


class WaveformPath(Base):
    """
    DB table containing file directories.
    """
    __tablename__ = 'default_waveform_paths'
    __table_args__ = (UniqueConstraint('path'), {})

    id = Column(Integer, primary_key=True)
    path = Column(String, nullable=False, index=True)
    archived = Column(Boolean, default=False)

    files = relation("WaveformFile", order_by="WaveformFile.id",
                     backref="path", cascade="all, delete, delete-orphan")

    def __init__(self, data={}):
        self.path = data.get('path')

    def __repr__(self):
        return "<WaveformPath('%s')>" % self.path


class WaveformFile(Base):
    """
    DB table containing waveform files.
    """
    __tablename__ = 'default_waveform_files'
    __table_args__ = (UniqueConstraint('file', 'path_id'), {})

    id = Column(Integer, primary_key=True)
    file = Column(String, nullable=False, index=True)
    size = Column(Integer, nullable=False)
    mtime = Column(Integer, nullable=False, index=True)
    format = Column(String, nullable=False, index=True)
    path_id = Column(Integer, ForeignKey('default_waveform_paths.id'))

    channels = relation("WaveformChannel", order_by="WaveformChannel.id",
                        backref="file", cascade="all, delete, delete-orphan")

    def __init__(self, data={}):
        self.file = data.get('file')
        self.size = data.get('size')
        self.mtime = int(data.get('mtime'))
        self.format = data.get('format')

    def __repr__(self):
        return "<WaveformFile('%s')>" % (self.id)


class WaveformChannel(Base):
    """
    DB table containing channels.
    """
    __tablename__ = 'default_waveform_channels'
    __table_args__ = (UniqueConstraint('network', 'station', 'location',
                                       'channel', 'file_id'), {})

    id = Column(Integer, primary_key=True)
    file_id = Column(Integer, ForeignKey('default_waveform_files.id'),
                     index=True)
    network = Column(String(2), nullable=False, index=True)
    station = Column(String(5), nullable=False, index=True)
    location = Column(String(2), nullable=False, index=True)
    channel = Column(String(3), nullable=False, index=True)
    starttime = Column(DateTime, nullable=False, index=True)
    endtime = Column(DateTime, nullable=False, index=True)
    calib = Column(Float, nullable=False)
    sampling_rate = Column(Float, nullable=False)
    npts = Column(Integer, nullable=False)
    preview = Column(PickleType(protocol=0), nullable=True)

    gaps = relation("WaveformGaps", order_by="WaveformGaps.id",
                    backref="channel", cascade="all, delete, delete-orphan")

    features = relation("WaveformFeatures", order_by="WaveformFeatures.id",
                        backref="channel",
                        cascade="all, delete, delete-orphan")

    def __init__(self, data={}):
        self.update(data)

    def update(self, data):
        self.network = data.get('network', '')
        self.station = data.get('station', '')
        self.location = data.get('location', '')
        self.channel = data.get('channel', '')
        self.starttime = data.get('starttime')
        self.endtime = data.get('endtime')
        self.calib = data.get('calib', 1.0)
        self.npts = data.get('npts', 0)
        self.sampling_rate = data.get('sampling_rate', 1.0)
        self.preview = data.get('preview', None)

    def __repr__(self):
        return "<WaveformChannel('%s')>" % (self.id)

    def get_preview(self, apply_calibration=False):
        try:
            data = pickle.loads(self.preview)
        except Exception:
            data = np.array([])
        if apply_calibration:
            data = data * self.calib
        tr = Trace(data=data)
        tr.stats.starttime = UTCDateTime(self.starttime)
        tr.stats.delta = 30.0
        tr.stats.network = self.network
        tr.stats.station = self.station
        tr.stats.location = self.location
        tr.stats.channel = self.channel
        tr.stats.calib = self.calib
        tr.stats.preview = True
        return tr


class WaveformGaps(Base):
    """
    DB table containing gaps.
    """
    __tablename__ = 'default_waveform_gaps'

    id = Column(Integer, primary_key=True)
    channel_id = Column(Integer, ForeignKey('default_waveform_channels.id'),
                        index=True)
    gap = Column(Boolean, nullable=False, index=True)
    starttime = Column(DateTime, nullable=False, index=True)
    endtime = Column(DateTime, nullable=False, index=True)
    samples = Column(Integer, nullable=False)

    def __init__(self, data={}):
        self.gap = data.get('gap', True)
        self.starttime = data.get('starttime')
        self.endtime = data.get('endtime')
        self.samples = data.get('samples', 0)

    def __repr__(self):
        return "<WaveformGaps('%s')>" % (self.id)


class WaveformFeatures(Base):
    """
    DB table containing optional features created during indexing.
    """
    __tablename__ = 'default_waveform_features'
    __table_args__ = (UniqueConstraint('channel_id', 'key'), {})

    id = Column(Integer, primary_key=True)
    channel_id = Column(Integer, ForeignKey('default_waveform_channels.id'),
                        index=True)
    key = Column(String, nullable=False, index=True)
    value = Column(PickleType, nullable=True)

    def __init__(self, data={}):
        self.key = data.get('key')
        self.value = pickle.dumps(data.get('value', None))

    def __repr__(self):
        return "<WaveformFeatures('%s')>" % (self.id)



# -*- coding: utf-8 -*-
"""
Library name handling for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
# NO IMPORTS FROM OBSPY OR FUTURE IN THIS FILE! (file gets used at
# installation time)
import ctypes
import os
import platform
import re
import warnings
from distutils import sysconfig


def cleanse_pymodule_filename(filename):
    """
    Replace all characters not allowed in Python module names in filename with
    "_".

    See bug report:
     - https://stackoverflow.com/q/21853678
     - See #755

    See also:
     - https://stackoverflow.com/q/7552311
     - https://docs.python.org/3/reference/lexical_analysis.html#identifiers

    >>> print(cleanse_pymodule_filename("0blup-bli.554_3!32"))
    _blup_bli_554_3_32
    """
    filename = re.sub(r'^[^a-zA-Z_]', "_", filename)
    filename = re.sub(r'[^a-zA-Z0-9_]', "_", filename)
    return filename


def _get_lib_name(lib, add_extension_suffix):
    """
    Helper function to get an architecture and Python version specific library
    filename.

    :type add_extension_suffix: bool
    :param add_extension_suffix: NumPy distutils adds a suffix to
        the filename we specify to build internally (as specified by Python
        builtin `sysconfig.get_config_var("EXT_SUFFIX")`. So when loading the
        file we have to add this suffix, but not during building.
    """
    # our custom defined part of the extension file name
    libname = "lib%s_%s_%s_py%s" % (
        lib, platform.system(), platform.architecture()[0],
        ''.join([str(i) for i in platform.python_version_tuple()[:2]]))
    libname = cleanse_pymodule_filename(libname)
    # NumPy distutils adds extension suffix by itself during build (#771, #755)
    if add_extension_suffix:
        # append any extension suffix defined by Python for current platform
        ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
        # in principle "EXT_SUFFIX" is what we want.
        # "SO" seems to be deprecated on newer python
        # but: older python seems to have empty "EXT_SUFFIX", so we fall back
        if not ext_suffix:
            try:
                ext_suffix = sysconfig.get_config_var("SO")
            except Exception as e:
                msg = ("Empty 'EXT_SUFFIX' encountered while building CDLL "
                       "filename and fallback to 'SO' variable failed "
                       "(%s)." % str(e))
                warnings.warn(msg)
                pass
        if ext_suffix:
            libname = libname + ext_suffix
    return libname


def _load_cdll(name):
    """
    Helper function to load a shared library built during ObsPy installation
    with ctypes.

    :type name: str
    :param name: Name of the library to load (e.g. 'mseed').
    :rtype: :class:`ctypes.CDLL`
    """
    # our custom defined part of the extension file name
    libname = _get_lib_name(name, add_extension_suffix=True)
    libdir = os.path.join(os.path.dirname(__file__), os.pardir, os.pardir,
                          'lib')
    libpath = os.path.join(libdir, libname)
    # resolve parent directory '../' for windows
    libpath = os.path.normpath(libpath)
    libpath = os.path.abspath(libpath)
    libpath = str(libpath)
    try:
        cdll = ctypes.CDLL(libpath)
    except Exception as e:
        import textwrap
        dirlisting = textwrap.wrap(
            ', '.join(sorted(os.path.dirname(libpath))))
        msg = ['Could not load shared library "%s"' % libname,
               'Path: %s' % libpath,
               'Current directory: %s' % os.path.abspath(os.curdir),
               'ctypes error message: %s' % str(e),
               'Directory listing of lib directory:',
               '  ',
               ]
        msg = '\n  '.join(msg)
        msg = msg + '\n    '.join(dirlisting)
        raise ImportError(msg)
    return cdll


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.io.kinemetrics.core test suite.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import io
import os
import unittest

import numpy as np

from obspy import read
from obspy.core.utcdatetime import UTCDateTime
from obspy.io.kinemetrics.core import is_evt, read_evt


class CoreTestCase(unittest.TestCase):
    """
    Test cases for kinemetrics core interface
    """
    def setUp(self):
        # directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data')

    def test_is_evt(self):
        """
        Test for the is_evt() function.
        """
        valid_files = [os.path.join(self.path, "BI008_MEMA-04823.evt"),
                       os.path.join(self.path, "BX456_MOLA-02351.evt")]
        invalid_files = [os.path.join(self.path, "NOUTF8.evt")]
        py_dir = os.path.join(self.path, os.pardir, os.pardir)
        for filename in os.listdir(py_dir):
            if filename.endswith(".py"):
                invalid_files.append(
                    os.path.abspath(os.path.join(py_dir, filename)))
        self.assertGreater(len(invalid_files), 0)

        for filename in valid_files:
            self.assertTrue(is_evt(filename))
        for filename in invalid_files:
            self.assertFalse(is_evt(filename))

    def test_is_evt_from_bytesio(self):
        """
        Test for the is_evt() function from BytesIO objects.
        """
        valid_files = [os.path.join(self.path, "BI008_MEMA-04823.evt"),
                       os.path.join(self.path, "BX456_MOLA-02351.evt")]
        invalid_files = [os.path.join(self.path, "NOUTF8.evt")]
        py_dir = os.path.join(self.path, os.pardir, os.pardir)
        for filename in os.listdir(py_dir):
            if filename.endswith(".py"):
                invalid_files.append(
                    os.path.abspath(os.path.join(py_dir, filename)))

        for filename in valid_files:
            with open(filename, "rb") as fh:
                buf = io.BytesIO(fh.read())
            buf.seek(0, 0)
            self.assertTrue(is_evt(buf))
            # The is_evt() method should not change the file pointer.
            self.assertEqual(buf.tell(), 0)
        for filename in invalid_files:
            with open(filename, "rb") as fh:
                buf = io.BytesIO(fh.read())
            buf.seek(0, 0)
            self.assertFalse(is_evt(buf))
            # The is_evt() method should not change the file pointer.
            self.assertEqual(buf.tell(), 0)

    def test_read_via_obspy(self):
        """
        Read files via obspy.core.stream.read function.
        """
        filename = os.path.join(self.path, 'BI008_MEMA-04823.evt')
        # 1
        st = read(filename)
        st.verify()
        self.assertEqual(len(st), 3)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(len(st[0]), 230 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MEMA')

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

        # 2
        filename = os.path.join(self.path, 'BX456_MOLA-02351.evt')
        st = read(filename)
        st.verify()
        self.assertEqual(len(st), 6)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[3].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[4].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[5].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(len(st[0]), 390 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MOLA')

    def test_reading_via_obspy_and_bytesio(self):
        """
        Test the reading of Evt files from BytesIO objects.
        """
        # 1
        filename = os.path.join(self.path, 'BI008_MEMA-04823.evt')
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read(buf)
        st.verify()
        self.assertEqual(len(st), 3)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(len(st[0]), 230 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MEMA')

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

        # 2
        filename = os.path.join(self.path, 'BX456_MOLA-02351.evt')
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read(buf)
        st.verify()
        self.assertEqual(len(st), 6)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[3].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[4].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[5].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(len(st[0]), 390 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MOLA')

    def test_read_via_module(self):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function.
        """
        filename = os.path.join(self.path, 'BI008_MEMA-04823.evt')
        # 1
        st = read_evt(filename)
        st.verify()
        self.assertEqual(len(st), 3)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(len(st[0]), 230 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MEMA')

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

        # 2
        filename = os.path.join(self.path, 'BX456_MOLA-02351.evt')
        st = read_evt(filename)
        st.verify()
        self.assertEqual(len(st), 6)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[3].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[4].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[5].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(len(st[0]), 390 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MOLA')

    def test_read_via_module_and_bytesio(self):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function from BytesIO
        objects.
        """
        # 1
        filename = os.path.join(self.path, 'BI008_MEMA-04823.evt')
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read_evt(buf)
        st.verify()
        self.assertEqual(len(st), 3)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2013-08-15T09:20:28.000000Z'))
        self.assertEqual(len(st[0]), 230 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MEMA')

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

        # 2
        filename = os.path.join(self.path, 'BX456_MOLA-02351.evt')
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read_evt(buf)
        st.verify()
        self.assertEqual(len(st), 6)
        self.assertEqual(st[0].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[1].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[2].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[3].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[4].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(st[5].stats.starttime,
                         UTCDateTime('2012-01-17T09:54:36.000000Z'))
        self.assertEqual(len(st[0]), 390 * 25)
        self.assertAlmostEqual(st[0].stats.sampling_rate, 250.0)
        self.assertEqual(st[0].stats.channel, '0')
        self.assertEqual(st[0].stats.station, 'MOLA')

    def verify_stats_evt(self, evt_stats):
        dico = {'chan_fullscale': 2.5, 'chan_sensorgain': 1,
                'chan_calcoil': 0.0500, 'chan_damping': 0.7070,
                'chan_natfreq': 196.00,
                'latitude': 50.609795, 'longitude': 6.009250,
                'elevation': 298}
        # Values from Kinemetrics QLWin program for BI008_MEMA-04823.evt

        for key in dico:
            self.assertAlmostEqual(dico[key], evt_stats[key], 6)

        self.assertEqual(UTCDateTime(2013, 8, 15, 9, 20, 28),
                         evt_stats['starttime'])

    def verify_data_evt0(self, data):
        valuesdeb = np.array([-2.4464752525e-002, -2.4534918368e-002,
                              -2.4467090145e-002, -2.4511529133e-002,
                              -2.4478785694e-002, -2.4483462796e-002,
                              -2.4434346706e-002, -2.4504512548e-002,
                              -2.4527901784e-002, -2.4455396459e-002,
                              -2.4509189650e-002, -2.4478785694e-002,
                              -2.4541934952e-002, -2.4497495964e-002,
                              -2.4448379874e-002, -2.4502173066e-002,
                              -2.4420313537e-002, -2.4455396459e-002,
                              -2.4546612054e-002, -2.4509189650e-002])
        valuesend = np.array([-2.4488139898e-002, -2.4530241266e-002,
                              -2.4525562301e-002, -2.4506852031e-002,
                              -2.4424990639e-002])
        # Data values from Tsoft Program

        self.assertEqual(len(data), 5750)
        self.assertTrue(np.allclose(valuesdeb, data[:len(valuesdeb)]))
        self.assertTrue(np.allclose(valuesend, data[-len(valuesend):]))

    def verify_data_evt2(self, data):
        valuesdeb = np.array([-4.4351171702e-002, -4.4479820877e-002,
                              -4.4447075576e-002, -4.4367544353e-002,
                              -4.4402632862e-002, -4.4386260211e-002,
                              -4.4360529631e-002, -4.4440057129e-002,
                              -4.4411987066e-002, -4.4395614415e-002,
                              -4.4421344995e-002, -4.4433038682e-002,
                              -4.4442396611e-002, -4.4423684478e-002,
                              -4.4428363442e-002, -4.4419005513e-002,
                              -4.4388595968e-002, -4.4360529631e-002,
                              -4.4358190149e-002, -4.4362869114e-002])
        valuesend = np.array([-4.4538296759e-002, -4.4549994171e-002,
                              -4.4493857771e-002, -4.4451754540e-002,
                              -4.4409647584e-002])

        # Data values from Tsoft Program
        # length is 5750

        self.assertEqual(len(data), 5750)
        self.assertTrue(np.allclose(valuesdeb, data[:len(valuesdeb)]))
        self.assertTrue(np.allclose(valuesend, data[-len(valuesend):]))


def suite():
    return unittest.makeSuite(CoreTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
obspy.clients.nrl - Nominal Response Library client for ObsPy
=============================================================

This module contains a client to access the Nominal Response Library (NRL)
hosted by IRIS (http://ds.iris.edu/NRL/).

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Basic Usage
-----------

The first step is to initialize a NRL client object. A client object can be
initialized either with the base URL of a NRL hosted on a web server or with a
local directory path to a downloaded and unpacked NRL zip file. The default is
to access the always up-to-date NRL database hosted at IRIS.

>>> from obspy.clients.nrl import NRL
>>> nrl = NRL()

The sensor and datalogger tree structure can be interactively explored in an
interactive Python shell:

>>> print(nrl) # doctest: +SKIP
NRL library at http://ds.iris.edu/NRL/
  Sensors: 24 manufacturers
    'CEA-DASE', 'CME', 'Chaparral Physics', 'Eentec', 'GEObit',
    'GEOsig', 'Generic', 'Geo Space/OYO', 'Geodevice', 'Geotech',
    'Guralp', 'Hyperion', 'IESE', 'Kinemetrics', 'LaHusen', 'Lennartz'
    'Metrozet', 'Nanometrics', 'REF TEK', 'Sercel/Mark Products',
    'Silicon Audio', 'SolGeo', 'Sprengnether (now Eentec)',
    'Streckeisen'
  Dataloggers: 15 manufacturers
    'Agecodagis', 'DAQ Systems (NetDAS)', 'Earth Data', 'Eentec',
    'Generic', 'GeoBIT', 'Geodevice', 'Geotech', 'Guralp',
    'Kinemetrics', 'Nanometrics', 'Omnirecs', 'Quanterra', 'REF TEK',
    'SolGeo'
>>> print(nrl.sensors) # doctest: +SKIP
Select the sensor manufacturer (24 items):
  'CEA-DASE', 'CME', 'Chaparral Physics', 'Eentec', 'GEObit', 'GEOsig'
  'Generic', 'Geo Space/OYO', 'Geodevice', 'Geotech', 'Guralp',
  'Hyperion', 'IESE', 'Kinemetrics', 'LaHusen', 'Lennartz', 'Metrozet'
  'Nanometrics', 'REF TEK', 'Sercel/Mark Products', 'Silicon Audio'
  'SolGeo', 'Sprengnether (now Eentec)', 'Streckeisen'
>>> print(nrl.sensors['Streckeisen']) # doctest: +SKIP
Select the Streckeisen sensor model (5 items):
  'STS-1', 'STS-2', 'STS-2.5', 'STS-3', 'STS-5A'
>>> print(nrl.sensors['Streckeisen']['STS-1']) # doctest: +SKIP
Select the corner period mode for this STS-1 (2 items):
  '20 seconds', '360 seconds'
>>> print(nrl.sensors['Streckeisen']['STS-1']['360 seconds']) # doctest: +SKIP
(u'STS-1, 360 s mode, 2400 V/m/s',
 u'http://ds.iris.edu/NRL/sensors/streckeisen/RESP.XX.NS088..BHZ.STS1.360.2400')  # NOQA

Response objects can be extracted by providing the datalogger and sensor keys:

>>> response = nrl.get_response( # doctest: +SKIP
...     sensor_keys=['Streckeisen', 'STS-1', '360 seconds'],
...     datalogger_keys=['REF TEK', 'RT 130 & 130-SMA', '1', '200'])
>>> print(response) # doctest: +SKIP
Channel Response
    From M/S (Velocity in Meters per Second) to COUNTS (Digital Counts)
    Overall Sensitivity: 1.50991e+09 defined at 0.020 Hz
    10 stages:
        Stage 1: PolesZerosResponseStage from M/S to V, gain: 2400
        Stage 2: ResponseStage from V to V, gain: 1
        Stage 3: CoefficientsTypeResponseStage from V to COUNTS, gain: 629129
        Stage 4: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 5: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 6: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 7: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 8: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 9: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
        Stage 10: CoefficientsTypeResponseStage from COUNTS to COUNTS, gain: 1
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

from .client import NRL



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import io
import os
import unittest
import datetime
import warnings
import random

import numpy as np

from obspy import UTCDateTime, read
from obspy.core.util import NamedTemporaryFile
from obspy.geodetics import gps2dist_azimuth, kilometer2degrees

from .. import header as _hd
from ..sactrace import SACTrace
from ..util import SacHeaderError, SacHeaderTimeError


class SACTraceTestCase(unittest.TestCase):
    """
    Test suite for obspy.io.sac.sactrace
    """
    def setUp(self):
        self.path = os.path.dirname(__file__)
        self.file = os.path.join(self.path, 'data', 'test.sac')
        self.filexy = os.path.join(self.path, 'data', 'testxy.sac')
        self.filebe = os.path.join(self.path, 'data', 'test.sac.swap')
        self.fileseis = os.path.join(self.path, 'data', 'seism.sac')
        self.testdata = np.array(
            [-8.74227766e-08, -3.09016973e-01,
             -5.87785363e-01, -8.09017122e-01, -9.51056600e-01,
             -1.00000000e+00, -9.51056302e-01, -8.09016585e-01,
             -5.87784529e-01, -3.09016049e-01], dtype=np.float32)

    def test_read_binary(self):
        """
        Tests for SACTrace binary file read
        """
        sac = SACTrace.read(self.file, byteorder='little')
        self.assertEqual(sac.npts, 100)
        self.assertEqual(sac.kstnm, 'STA')
        self.assertEqual(sac.delta, 1.0)
        self.assertEqual(sac.kcmpnm, 'Q')
        self.assertEqual(sac.reftime.datetime,
                         datetime.datetime(1978, 7, 18, 8, 0))
        self.assertEqual(sac.nvhdr, 6)
        self.assertEqual(sac.b, 10.0)
        self.assertAlmostEqual(sac.depmen, 9.0599059e-8)
        np.testing.assert_array_almost_equal(self.testdata[0:10],
                                             sac.data[0:10])

    def test_read_binary_headonly(self):
        """
        A headonly read should return readable headers and data is None
        """
        sac = SACTrace.read(self.file, byteorder='little', headonly=True)
        self.assertEqual(sac.data, None)
        self.assertEqual(sac.npts, 100)
        self.assertEqual(sac.depmin, -1.0)
        self.assertAlmostEqual(sac.depmen, 8.344650e-8)
        self.assertEqual(sac.depmax, 1.0)

    def test_read_sac_byteorder(self):
        """
        A read should fail if the byteorder is wrong
        """
        with self.assertRaises(IOError):
            SACTrace.read(self.filebe, byteorder='little')
        with self.assertRaises(IOError):
            SACTrace.read(self.file, byteorder='big')
        # a SACTrace should show the correct byteorder
        sac = SACTrace.read(self.filebe, byteorder='big')
        self.assertEqual(sac.byteorder, 'big')
        sac = SACTrace.read(self.file, byteorder='little')
        self.assertEqual(sac.byteorder, 'little')
        # a SACTrace should autodetect the correct byteorder
        sac = SACTrace.read(self.file)
        self.assertEqual(sac.byteorder, 'little')
        sac = SACTrace.read(self.filebe)
        self.assertEqual(sac.byteorder, 'big')

    def test_write_sac(self):
        """
        A trace you've written and read in again should look the same as the
        one you started with.
        """
        sac1 = SACTrace.read(self.file, byteorder='little')
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            sac1.write(tempfile, byteorder='little')
            sac2 = SACTrace.read(tempfile, byteorder='little')
        np.testing.assert_array_equal(sac1.data, sac2.data)
        self.assertEqual(sac1._header, sac2._header)

    def test_write_binary_headonly(self):
        """
        A trace you've written headonly should only modify the header of an
        existing file, and fail if the file doesn't exist.
        """
        # make a sac trace
        sac = SACTrace.read(self.file, byteorder='little')
        # write it all to temp file
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            sac.write(tempfile, byteorder='little')
            # read it headonly and modify the header
            # modify the data, too, and verify it didn't get written
            sac2 = SACTrace.read(tempfile, headonly=True, byteorder='little')
            sac2.kcmpnm = 'xyz'
            sac2.b = 7.5
            sac2.data = np.array([1.5, 2e-3, 17], dtype=np.float32)
            # write it again (write over)
            sac2.write(tempfile, headonly=True, byteorder='little')
            # read it all and compare
            sac3 = SACTrace.read(tempfile, byteorder='little')
        self.assertEqual(sac3.kcmpnm, 'xyz')
        self.assertEqual(sac3.b, 7.5)
        np.testing.assert_array_equal(sac3.data, sac.data)

        # ...and fail if the file doesn't exist
        sac = SACTrace.read(self.file, headonly=True, byteorder='little')
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
        with self.assertRaises(IOError):
            sac.write(tempfile, headonly=True, byteorder='little')

    def test_read_sac_ascii(self):
        """
        Read an ASCII SAC file.
        """
        sac = SACTrace.read(self.filexy, ascii=True)
        self.assertEqual(sac.npts, 100)
        self.assertEqual(sac.kstnm, 'sta')
        self.assertEqual(sac.delta, 1.0)
        self.assertEqual(sac.kcmpnm, 'Q')
        self.assertEqual(sac.nvhdr, 6)
        self.assertEqual(sac.b, 10.0)
        self.assertAlmostEqual(sac.depmen, 9.4771387e-08)
        np.testing.assert_array_almost_equal(self.testdata[0:10],
                                             sac.data[0:10])

    def test_reftime(self):
        """
        A SACTrace.reftime should be created correctly from a file's nz-times
        """
        sac = SACTrace.read(self.fileseis)
        self.assertEqual(sac.reftime,
                         UTCDateTime('1981-03-29T10:38:14.000000Z'))
        # changes to a reftime should be reflected in the nz times and reftime
        nzsec, nzmsec = sac.nzsec, sac.nzmsec
        sac.reftime = sac.reftime + 2.5
        self.assertEqual(sac.nzsec, nzsec + 2)
        self.assertEqual(sac.nzmsec, nzmsec + 500)
        self.assertEqual(sac.reftime,
                         UTCDateTime('1981-03-29T10:38:16.500000Z'))
        # changes in the nztimes should be reflected reftime
        sac.nzyear = 2001
        self.assertEqual(sac.reftime.year, 2001)

    def test_reftime_relative_times(self):
        """
        Changes in the reftime shift all relative time headers
        """
        sac = SACTrace.read(self.fileseis)
        a, b, t1 = sac.a, sac.b, sac.t1
        sac.reftime -= 10.0
        self.assertAlmostEqual(sac.a, a + 10.0, 5)
        self.assertAlmostEqual(sac.b, b + 10.0)
        self.assertAlmostEqual(sac.t1, t1 + 10.0)
        # changes in the reftime should push remainder microseconds to the
        # relative time headers, and milliseconds to the nzmsec
        sac = SACTrace(b=5.0, t1=20.0)
        b, t1, nzmsec = sac.b, sac.t1, sac.nzmsec
        sac.reftime += 1.2e-3
        self.assertEqual(sac.nzmsec, nzmsec + 1)
        self.assertAlmostEqual(sac.b, b - 1.0e-3, 6)
        self.assertAlmostEqual(sac.t1, t1 - 1.0e-3, 5)

    def test_lcalda(self):
        """
        Test that distances are set when geographic information is present and
        lcalda is True, and that they're not set when geographic information
        is missing or lcalca is false.
        """
        stla, stlo, evla, evlo = -35.0, 100, 42.5, -37.5
        meters, az, baz = gps2dist_azimuth(evla, evlo, stla, stlo)
        km = meters / 1000.0
        gcarc = kilometer2degrees(km)

        # distances are set when lcalda True and all distance values set
        sac = SACTrace(lcalda=True, stla=stla, stlo=stlo, evla=evla, evlo=evlo)
        self.assertAlmostEqual(sac.az, az, places=4)
        self.assertAlmostEqual(sac.baz, baz, places=4)
        self.assertAlmostEqual(sac.dist, km, places=4)
        self.assertAlmostEqual(sac.gcarc, gcarc, places=4)
        # distances are not set when lcalda False and all distance values set
        sac = SACTrace(lcalda=False, stla=stla, stlo=stlo, evla=evla,
                       evlo=evlo)
        self.assertIs(sac.az, None)
        self.assertIs(sac.baz, None)
        self.assertIs(sac.dist, None)
        self.assertIs(sac.gcarc, None)
        # distances are not set when lcalda True, not all distance values set
        sac = SACTrace(lcalda=True, stla=stla)
        self.assertIs(sac.az, None)
        self.assertIs(sac.baz, None)
        self.assertIs(sac.dist, None)
        self.assertIs(sac.gcarc, None)
        # exception raised when set_distances is forced but not all distances
        # values are set. NOTE: still have a problem when others are "None".
        sac = SACTrace(lcalda=True, stla=stla)
        self.assertRaises(SacHeaderError, sac._set_distances, force=True)

    def test_propagate_modified_stats_strings_to_sactrace(self):
        """
        If you build a SACTrace from an ObsPy Trace that has certain string
        headers mismatched between the Stats header and an existing Stats.sac
        header, channel and kcmpnm for example, the resulting SACTrace values
        should come from Stats. Addresses GitHub issue #1457.
        """
        tr = read(self.fileseis)[0]
        # modify the header values by adding a single meaningless character
        for sachdr, statshdr in [('kstnm', 'station'), ('knetwk', 'network'),
                                 ('kcmpnm', 'channel'), ('khole', 'location')]:
            modified_value = tr.stats[statshdr] + '1'
            tr.stats[statshdr] = modified_value
            sac = SACTrace.from_obspy_trace(tr)
            self.assertEqual(getattr(sac, sachdr), modified_value)

    def test_reftime_incomplete(self):
        """
        Replacement for SACTrace._from_arrays doctest which raises UserWarning
        """
        sac = SACTrace._from_arrays()
        self.assertTrue(sac.lcalda)
        self.assertFalse(sac.leven)
        self.assertFalse(sac.lovrok)
        self.assertFalse(sac.lpspol)
        self.assertEqual(sac.iztype, None)
        self.assertRaises(SacHeaderTimeError, getattr, sac, 'reftime')
        # raises "UserWarning: Reference time information incomplete"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            str(sac)
            self.assertEqual(len(w), 1)
            self.assertIn("Reference time information incomplete", str(w[0]))

    def test_floatheader(self):
        """
        Test standard SACTrace float headers using the floatheader descriptor.
        """
        sac = SACTrace()
        for hdr in ('delta', 'scale', 'odelta', 'internal0', 'stel', 'stdp',
                    'evdp', 'mag', 'user0', 'user1', 'user2', 'user3', 'user4',
                    'user5', 'user6', 'user7', 'user8', 'user9', 'dist', 'az',
                    'baz', 'gcarc', 'cmpaz', 'cmpinc'):
            floatval = random.random()

            # setting value
            setattr(sac, hdr, floatval)
            self.assertAlmostEqual(sac._hf[_hd.FLOATHDRS.index(hdr)], floatval)
            # getting value
            self.assertAlmostEqual(getattr(sac, hdr), floatval)
            # setting None produces null value
            setattr(sac, hdr, None)
            self.assertAlmostEqual(sac._hf[_hd.FLOATHDRS.index(hdr)],
                                   _hd.FNULL)
            # getting existing null values return None
            sac._hf[_hd.FLOATHDRS.index(hdr)] = _hd.FNULL
            self.assertIsNone(getattr(sac, hdr))
            # __doc__ on class and instance
            self.assertEqual(getattr(SACTrace, hdr).__doc__, _hd.DOC.get(hdr))
            # self.assertEqual(getattr(sac, hdr).__doc__, _hd.DOC.get(hdr]))
            # TODO: I'd like to find a way for this to work:-(
            # TODO: factor __doc__ tests out into one test for all headers

    def test_relative_time_headers(self):
        """
        Setting relative time headers will work with UTCDateTime objects.
        """
        # TODO: ultimately, _all_ children of floatheader (this one, geosetter,
        #   etc.) should be tested in test_floatheader for normal setting, and
        #   only special behaviour will happen here.
        utc = UTCDateTime(year=1970, month=1, day=1, minute=15, second=10,
                          microsecond=0)
        sac = SACTrace(nzyear=utc.year, nzjday=utc.julday, nzhour=utc.hour,
                       nzmin=utc.minute, nzsec=utc.second, nzmsec=0)
        for hdr in ('b', 'a', 'o', 'f', 't0', 't1', 't2', 't3', 't4', 't5',
                    't6', 't7', 't8', 't9'):
            offset_float = random.uniform(-1, 1)
            offset_utc = utc + offset_float
            setattr(sac, hdr, offset_utc)
            self.assertAlmostEqual(sac._hf[_hd.FLOATHDRS.index(hdr)],
                                   offset_float, places=5)

    def test_int_headers(self):
        """
        Test standard SACTrace int headers using the intheader descriptor.
        """
        sac = SACTrace()
        for hdr in ('nzyear', 'nzjday', 'nzhour', 'nzmin', 'nzsec', 'nzmsec',
                    'nvhdr', 'norid', 'nevid', 'nwfid', 'iinst', 'istreg',
                    'ievreg', 'iqual', 'unused23'):

            intval = random.randint(-10, 10)

            # setting value
            setattr(sac, hdr, intval)
            self.assertEqual(sac._hi[_hd.INTHDRS.index(hdr)], intval)
            # getting value
            self.assertEqual(getattr(sac, hdr), intval)
            # setting None produces null value
            setattr(sac, hdr, None)
            self.assertEqual(sac._hi[_hd.INTHDRS.index(hdr)], _hd.INULL)
            # getting existing null values return None
            sac._hi[_hd.INTHDRS.index(hdr)] = _hd.INULL
            self.assertIsNone(getattr(sac, hdr))
            # __doc__ on class and instance
            self.assertEqual(getattr(SACTrace, hdr).__doc__, _hd.DOC.get(hdr))

    def test_bool_headers(self):
        sac = SACTrace()
        for hdr in ('leven', 'lpspol', 'lovrok', 'lcalda'):
            # getting existing null values return None
            sac._hi[_hd.INTHDRS.index(hdr)] = _hd.INULL
            self.assertIsNone(getattr(sac, hdr))

            for boolval in (True, False, 0, 1):
                setattr(sac, hdr, boolval)
                self.assertEqual(sac._hi[_hd.INTHDRS.index(hdr)], int(boolval))
                self.assertEqual(getattr(sac, hdr), bool(boolval))

    def test_enumheader(self):
        sac = SACTrace()
        # set all the `iztype` reference headers, so that it won't fail
        for idx, hdr in enumerate(['b', 'o', 'a', 'f'] +
                                  ['t' + str(i) for i in range(10)]):
            setattr(sac, hdr, idx)
        for enumhdr, accepted_vals in _hd.ACCEPTED_VALS.items():
            if enumhdr != 'iqual':
                # iqual is allowed to be integer, not an enumerated str
                for accepted_val in accepted_vals:
                    accepted_int = _hd.ENUM_VALS[accepted_val]

                    sac._hi[_hd.INTHDRS.index(enumhdr)] = accepted_int
                    self.assertEqual(getattr(sac, enumhdr), accepted_val)

                    setattr(sac, enumhdr, accepted_val)
                    self.assertEqual(sac._hi[_hd.INTHDRS.index(enumhdr)],
                                     accepted_int)

    def test_string_headers(self):
        sac = SACTrace()
        for hdr in ('kstnm', 'khole', 'ko', 'ka', 'kt0', 'kt1', 'kt2', 'kt3',
                    'kt4', 'kt5', 'kt6', 'kt7', 'kt8', 'kt9', 'kf', 'kuser0',
                    'kuser1', 'kuser2', 'kcmpnm', 'knetwk', 'kdatrd',
                    'kinst'):

            strval = hdr

            # normal get/set
            setattr(sac, hdr, strval)
            self.assertEqual(sac._hs[_hd.STRHDRS.index(hdr)].decode(), strval)
            self.assertEqual(getattr(sac, hdr), strval)

            # null get/set
            sac._hs[_hd.STRHDRS.index(hdr)] = _hd.SNULL
            self.assertIsNone(getattr(sac, hdr))

            # get/set value too long
            too_long = "{}_1234567890".format(hdr)
            # will raise "UserWarning: Alphanumeric headers longer than 8
            # characters are right-truncated"
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('always', UserWarning)
                setattr(sac, hdr, too_long)
                self.assertEqual(len(w), 1)
                self.assertEqual(w[0].category, UserWarning)
                self.assertIn('Alphanumeric headers longer than 8', str(w[0]))
            self.assertEqual(sac._hs[_hd.STRHDRS.index(hdr)].decode(),
                             too_long[:8])
            self.assertEqual(getattr(sac, hdr), too_long[:8].strip())

            # docstring
            self.assertEqual(getattr(SACTrace, hdr).__doc__, _hd.DOC.get(hdr))

    def test_kevnm(self):
        sac = SACTrace()
        # test kevnm (kevnm + kevnm2)
        kevnm = '1234567890123456'
        kevnm1, kevnm2 = kevnm[:8], kevnm[8:]

        sac.kevnm = kevnm
        self.assertEqual(sac._hs[_hd.STRHDRS.index('kevnm')].decode(), kevnm1)
        self.assertEqual(sac._hs[_hd.STRHDRS.index('kevnm2')].decode(), kevnm2)
        self.assertEqual(sac.kevnm, kevnm)

        sac._hs[_hd.STRHDRS.index('kevnm')] = _hd.SNULL
        sac._hs[_hd.STRHDRS.index('kevnm2')] = _hd.SNULL
        self.assertIsNone(sac.kevnm)

        self.assertEqual(SACTrace.kevnm.__doc__, _hd.DOC.get('kevnm'))

    def test_data_headers(self):
        """
        Headers that depend on the data vector should return values operating
        on the data, or fall back to stored header values if data is absent.
        """
        data = np.random.ranf(10).astype(np.float32)
        npts = 4
        depmax = 3.0
        depmen = 2.0
        depmin = 1.0
        sac = SACTrace(depmin=depmin, depmen=depmen, depmax=depmax, npts=npts,
                       data=data)

        for hdr, func in [('depmin', min), ('depmen', np.mean),
                          ('depmax', max), ('npts', len)]:
            # getting value
            self.assertEqual(getattr(sac, hdr), func(data))

            with self.assertRaises(AttributeError):
                # can't set value on write-only attribute
                setattr(sac, hdr, func(data))

        # headers fall back to stored value when data is None
        sac.data = None
        for hdr, value in [('depmin', depmin), ('depmen', depmen),
                           ('depmax', depmax), ('npts', npts)]:
            self.assertEqual(getattr(sac, hdr), value)

    def test_char_header_padding(self):
        """
        SAC binary file header fields should be padded with blanks, not NULs,
        or NUL-terminated.
        """

        dat = np.zeros(8)

        sac = SACTrace(
            kstnm='TEST',
            delta=1,
            kcmpnm='Z',
            kevnm='Bug test',
            npts=len(dat),
            data=dat
        )
        with io.BytesIO() as bio:
            sac.write(bio, byteorder='big')
            bio.seek(0)
            result = bio.read()
        # kstnm is at offset 0x1b8 in the file
        self.assertEqual(result[0x1b8:(0x1b8+8)], b'TEST    ')
        # kcmpnm is at offset 0x258 in the file
        self.assertEqual(result[0x258:(0x258+8)], b'Z       ')


def suite():
    return unittest.makeSuite(SACTraceTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

from .blockette import Blockette
from ..fields import Integer, VariableString


class Blockette034(Blockette):
    """
    Blockette 034: Units Abbreviations Blockette.

    This blockette defines the units of measurement in a standard, repeatable
    way. Mention each unit of measurement only once.

    Sample:
    0340044001M/S~Velocity in Meters Per Second~
    """
    id = 34
    name = "Units Abbreviations"
    fields = [
        Integer(3, "Unit lookup code", 3),
        VariableString(4, "Unit name", 1, 20, 'UNP'),
        VariableString(5, "Unit description", 0, 50, 'UNLPS')
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import inspect
import io
import os
import unittest

import obspy
from obspy.core.util.base import NamedTemporaryFile
from obspy.io.cmtsolution.core import _is_cmtsolution


class CmtsolutionTestCase(unittest.TestCase):
    """
    Test suite for obspy.io.cmtsolution.

    The tests usually directly utilize the registered function with the
    read_events() to also test the integration.
    """
    def setUp(self):
        self.path = os.path.dirname(os.path.abspath(inspect.getfile(
            inspect.currentframe())))
        self.datapath = os.path.join(self.path, "data")

    def test_read_and_write_cmtsolution_from_files(self):
        """
        Tests that reading and writing a CMTSOLUTION file does not change
        anything.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION")
        with open(filename, "rb") as fh:
            data = fh.read()

        cat = obspy.read_events(filename)

        with NamedTemporaryFile() as tf:
            temp_filename = tf.name

        try:
            cat.write(temp_filename, format="CMTSOLUTION")
            with open(temp_filename, "rb") as fh:
                new_data = fh.read()
        finally:
            try:
                os.remove(temp_filename)
            except Exception:
                pass

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_write_no_preferred_focal_mechanism(self):
        """
        Tests that writing a CMTSOLUTION file with no preferred (but at least
        one) focal mechanism works, see #1303.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION")
        with open(filename, "rb") as fh:
            data = fh.read()

        cat = obspy.read_events(filename)
        cat[0].preferred_focal_mechanism_id = None

        with NamedTemporaryFile() as tf:
            temp_filename = tf.name

        try:
            cat.write(temp_filename, format="CMTSOLUTION")
            with open(temp_filename, "rb") as fh:
                new_data = fh.read()
        finally:
            try:
                os.remove(temp_filename)
            except Exception:
                pass

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_read_and_write_cmtsolution_from_open_files(self):
        """
        Tests that reading and writing a CMTSOLUTION file does not change
        anything.

        This time it tests reading from and writing to open files.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION")
        with open(filename, "rb") as fh:
            data = fh.read()
            fh.seek(0, 0)
            cat = obspy.read_events(fh)

        with NamedTemporaryFile() as tf:
            cat.write(tf, format="CMTSOLUTION")
            tf.seek(0, 0)
            new_data = tf.read()

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_read_and_write_cmtsolution_from_bytes_io(self):
        """
        Tests that reading and writing a CMTSOLUTION file does not change
        anything.

        This time it tests reading from and writing to BytesIO objects.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION")
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
            data = buf.read()
            buf.seek(0, 0)

        with buf:
            buf.seek(0, 0)
            cat = obspy.read_events(buf)

            with io.BytesIO() as buf2:
                cat.write(buf2, format="CMTSOLUTION")
                buf2.seek(0, 0)
                new_data = buf2.read()

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_read_and_write_cmtsolution_explosion(self):
        """
        Tests that reading and writing a CMTSOLUTION file does not change
        anything.

        Tests another file.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION_EXPLOSION")
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())

        data = buf.read()
        buf.seek(0, 0)

        with buf:
            cat = obspy.read_events(buf)

            with io.BytesIO() as buf2:
                cat.write(buf2, format="CMTSOLUTION")
                buf2.seek(0, 0)
                new_data = buf2.read()

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_is_cmtsolution(self):
        """
        Tests the is_cmtsolution function.
        """
        good_files = [os.path.join(self.datapath, "CMTSOLUTION"),
                      os.path.join(self.datapath, "CMTSOLUTION_EXPLOSION")]

        bad_files = [
            os.path.join(self.datapath, os.path.pardir, "test_core.py"),
            os.path.join(self.datapath, os.path.pardir, "__init__.py")]

        for filename in good_files:
            self.assertTrue(_is_cmtsolution(filename))
        for filename in bad_files:
            self.assertFalse(_is_cmtsolution(filename))

    def test_read_and_write_multiple_cmtsolution_from_files(self):
        """
        Tests that reading and writing a CMTSOLUTION file with multiple
        events does not change anything.
        """
        filename = os.path.join(self.datapath, "MULTIPLE_EVENTS")
        with open(filename, "rb") as fh:
            data = fh.read()

        cat = obspy.read_events(filename)

        self.assertEqual(len(cat), 4)

        with NamedTemporaryFile() as tf:
            temp_filename = tf.name

        try:
            cat.write(temp_filename, format="CMTSOLUTION")
            with open(temp_filename, "rb") as fh:
                new_data = fh.read()
        finally:
            try:
                os.remove(temp_filename)
            except Exception:
                pass

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_read_and_write_multiple_events_from_bytes_io(self):
        """
        Tests that reading and writing a CMTSOLUTION file with multiple
        events does not change anything.

        This time it tests reading from and writing to BytesIO objects.
        """
        filename = os.path.join(self.datapath, "MULTIPLE_EVENTS")
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
            data = buf.read()
            buf.seek(0, 0)

        with buf:
            buf.seek(0, 0)
            cat = obspy.read_events(buf)

            self.assertEqual(len(cat), 4)

            with io.BytesIO() as buf2:
                cat.write(buf2, format="CMTSOLUTION")
                buf2.seek(0, 0)
                new_data = buf2.read()

        self.assertEqual(data.decode().splitlines(),
                         new_data.decode().splitlines())

    def test_reading_newer_cmtsolution_files(self):
        """
        The format changed a bit. Make sure these files can also be read.
        """
        filename = os.path.join(self.datapath, "CMTSOLUTION_NEW")
        cat = obspy.read_events(filename)

        self.assertEqual(len(cat), 3)

        # Test the hypocentral origins as the "change" to the format only
        # affected the first line.
        self.assertEqual(cat[0].origins[1].latitude, 55.29)
        self.assertEqual(cat[0].origins[1].longitude, 163.06)

        self.assertEqual(cat[1].origins[1].latitude, -13.75)
        self.assertEqual(cat[1].origins[1].longitude, -111.75)

        self.assertEqual(cat[2].origins[1].latitude, -13.68)
        self.assertEqual(cat[2].origins[1].longitude, -111.93)


def suite():
    return unittest.makeSuite(CmtsolutionTestCase, "test")


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# -*- coding: utf-8 -*-
"""
ObsPy's compatibility layer.

Includes things to easy dealing with Py2/Py3 differences as well as making
it work with various versions of our dependencies.
"""
from future.utils import PY2

import collections
import importlib
import io
import json
import sys
import unittest

import numpy as np

# optional dependencies
try:
    if PY2:
        import mock  # NOQA
    else:
        from unittest import mock  # NOQA
except ImportError:
    pass

if PY2:
    from string import maketrans  # NOQA
    from urlparse import urlparse  # NOQA
else:
    maketrans = bytes.maketrans  # NOQA
    from urllib.parse import urlparse  # NOQA


# Define the string types.
if PY2:
    string_types = (basestring,)  # NOQA
else:
    string_types = (str,)  # NOQA


# Importing the ABCs from collections will no longer work with Python 3.8.
if PY2:
    collections_abc = collections  # NOQA
else:
    try:
        collections_abc = collections.abc  # NOQA
    except AttributeError:
        # Python 3.4 compat, see https://bugs.python.org/msg212284
        # some older Linux distribution (like Debian jessie) are still in LTS,
        # so be nice, this doesn't hurt and can be removed again later on
        collections_abc = importlib.import_module("collections.abc")  # NOQA


if PY2:
    class RegExTestCase(unittest.TestCase):
        def assertRaisesRegex(self, exception, regex, callable,  # NOQA
                              *args, **kwargs):
            return self.assertRaisesRegexp(exception, regex, callable,
                                           *args, **kwargs)
else:
    class RegExTestCase(unittest.TestCase):
        pass


# NumPy does not offer the from_buffer method under Python 3 and instead
# relies on the built-in memoryview object.
if PY2:
    def from_buffer(data, dtype):
        # For compatibility with NumPy 1.4
        if isinstance(dtype, unicode):  # noqa
            dtype = str(dtype)
        if data:
            try:
                data = data.encode()
            except Exception:
                pass
            return np.frombuffer(data, dtype=dtype).copy()
        else:
            return np.array([], dtype=dtype)
else:
    def from_buffer(data, dtype):
        try:
            data = data.encode()
        except Exception:
            pass
        return np.array(memoryview(data)).view(dtype).copy()  # NOQA


if PY2:
    from ConfigParser import SafeConfigParser as ConfigParser  # NOQA
else:
    from configparser import ConfigParser  # NOQA


def is_text_buffer(obj):
    """
    Helper function determining if the passed object is an object that can
    read and write text or not.

    :param obj: The object to be tested.
    :return: True/False
    """
    # Default open()'ed files and StringIO (in Python 2) don't inherit from any
    # of the io classes thus we only test the methods of the objects which
    # in Python 2 should be safe enough.
    if PY2 and not isinstance(obj, io.BufferedIOBase) and \
            not isinstance(obj, io.TextIOBase):
        if hasattr(obj, "read") and hasattr(obj, "write") \
                and hasattr(obj, "seek") and hasattr(obj, "tell"):
            return True
        return False

    return isinstance(obj, io.TextIOBase)


def is_bytes_buffer(obj):
    """
    Helper function determining if the passed object is an object that can
    read and write bytes or not.

    :param obj: The object to be tested.
    :return: True/False
    """
    # Default open()'ed files and StringIO (in Python 2) don't inherit from any
    # of the io classes thus we only test the methods of the objects which
    # in Python 2 should be safe enough.
    if PY2 and not isinstance(obj, io.BufferedIOBase) and \
            not isinstance(obj, io.TextIOBase):
        if hasattr(obj, "read") and hasattr(obj, "write") \
                and hasattr(obj, "seek") and hasattr(obj, "tell"):
            return True
        return False

    return isinstance(obj, io.BufferedIOBase)


def round_away(number):
    """
    Simple function that rounds a number to the nearest integer. If the number
    is halfway between two integers, it will round away from zero. Of course
    only works up machine precision. This should hopefully behave like the
    round() function in Python 2.

    This is potentially desired behavior in the trim functions but some more
    thought should be poured into it.

    The np.round() function rounds towards the even nearest even number in case
    of half-way splits.

    >>> round_away(2.5)
    3
    >>> round_away(-2.5)
    -3

    >>> round_away(10.5)
    11
    >>> round_away(-10.5)
    -11

    >>> round_away(11.0)
    11
    >>> round_away(-11.0)
    -11
    """
    floor = np.floor(number)
    ceil = np.ceil(number)
    if (floor != ceil) and (abs(number - floor) == abs(ceil - number)):
        return int(int(number) + int(np.sign(number)))
    else:
        return int(np.round(number))


if sys.version_info[0] < 3:
    def py3_round(number, ndigits=None):
        """
        Similar function to python 3's built-in round function.

        Returns a float if ndigits is greater than 0, else returns an integer.

        Note:
        This function should be replace by the builtin round when obspy
        drops support for python 2.7.
        Unlike python'3 rounding, this function always rounds up on half
        increments rather than implementing banker's rounding.

        :type number: int or float
        :param number: A real number to be rounded
        :type ndigits: int
        :param ndigits: number of digits
        :return: An int if ndigites <= 0, else a float rounded to ndigits.
        """
        if ndigits is None or ndigits <= 0:
            mult = 10 ** -(ndigits or 0)
            return ((int(number) + mult // 2) // mult) * mult
        else:
            return round(number, ndigits)
else:
    py3_round = round


def get_json_from_response(r):
    """
    Get a JSON response in a way that also works for very old request
    versions.

    :type r: :class:`requests.Response
    :param r: The server's response.
    """
    if hasattr(r, "json"):
        if isinstance(r.json, dict):
            return r.json
        return r.json()

    c = r.content
    try:
        c = c.decode()
    except Exception:
        pass
    return json.loads(c)


def get_text_from_response(r):
    """
    Get a text response in a way that also works for very old request versions.

    :type r: :class:`requests.Response
    :param r: The server's response.
    """
    if hasattr(r, "text"):
        return r.text

    c = r.content
    try:
        c = c.decode()
    except Exception:
        pass
    return c


def get_reason_from_response(r):
    """
    Get the status text.

    :type r: :class:`requests.Response
    :param r: The server's response.
    """
    # Very old requests version might not have the reason attribute.
    if hasattr(r, "reason"):
        c = r.reason
    else:  # pragma: no cover
        c = r.raw.reason

    if hasattr(c, "encode"):
        c = c.encode()

    return c



# -*- coding: utf-8 -*-
"""
AH bindings to ObsPy core module.

An AH file is used for the storage of binary seismic time series data.
The file is portable among machines of varying architecture by virtue of
its XDR implementation. It is composed of a variable-sized header containing
a number of values followed by the time series data.

.. seealso:: ftp://www.orfeus-eu.org/pub/software/mirror/ldeo.columbia/

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import xdrlib
import numpy as np

from obspy import Stream, Trace, UTCDateTime
from obspy.core.util.attribdict import AttribDict

AH1_CODESIZE = 6
AH1_CHANSIZE = 6
AH1_STYPESIZE = 8
AH1_COMSIZE = 80
AH1_LOGSIZE = 202


def _is_ah(filename):
    """
    Checks whether a file is AH waveform data or not.

    :type filename: str
    :param filename: AH file to be checked.
    :rtype: bool
    :return: ``True`` if a AH waveform file.
    """
    if _get_ah_version(filename):
        return True
    return False


def _read_ah(filename, **kwargs):  # @UnusedVariable
    """
    Reads an AH waveform file and returns a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type filename: str
    :param filename: AH file to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream with Traces specified by given file.
    """
    version = _get_ah_version(filename)
    if version == '2.0':
        return _read_ah2(filename)
    else:
        return _read_ah1(filename)


def _get_ah_version(filename):
    """
    Returns version of AH waveform data.

    :type filename: str
    :param filename: AH v1 file to be checked.
    :rtype: str or False
    :return: version string of AH waveform data or ``False`` if unknown.
    """
    with open(filename, "rb") as fh:
        # read first 8 bytes with XDR library
        try:
            data = xdrlib.Unpacker(fh.read(8))
            # check for magic version number
            magic = data.unpack_int()
        except Exception:
            return False
        if magic == 1100:
            try:
                # get record length
                length = data.unpack_uint()
                # read first record
                fh.read(length)
            except Exception:
                return False
            # seems to be AH v2
            return '2.0'
        elif magic == 6:
            # AH1 has no magic variable :/
            # so we have to use some fixed values as indicators
            try:
                fh.seek(12)
                if xdrlib.Unpacker(fh.read(4)).unpack_int() != 6:
                    return False
                fh.seek(24)
                if xdrlib.Unpacker(fh.read(4)).unpack_int() != 8:
                    return False
                fh.seek(700)
                if xdrlib.Unpacker(fh.read(4)).unpack_int() != 80:
                    return False
                fh.seek(784)
                if xdrlib.Unpacker(fh.read(4)).unpack_int() != 202:
                    return False
            except Exception:
                return False
            return '1.0'
        else:
            return False


def _unpack_string(data):
    return data.unpack_string().split(b'\x00', 1)[0].strip().decode("utf-8")


def _read_ah1(filename):
    """
    Reads an AH v1 waveform file and returns a Stream object.

    :type filename: str
    :param filename: AH v1 file to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream with Traces specified by given file.
    """

    def _unpack_trace(data):
        ah_stats = AttribDict({
            'version': '1.0',
            'event': AttribDict(),
            'station': AttribDict(),
            'record': AttribDict(),
            'extras': []
        })

        # station info
        ah_stats.station.code = _unpack_string(data)
        ah_stats.station.channel = _unpack_string(data)
        ah_stats.station.type = _unpack_string(data)
        ah_stats.station.latitude = data.unpack_float()
        ah_stats.station.longitude = data.unpack_float()
        ah_stats.station.elevation = data.unpack_float()
        ah_stats.station.gain = data.unpack_float()
        ah_stats.station.normalization = data.unpack_float()  # A0
        poles = []
        zeros = []
        for _i in range(0, 30):
            r = data.unpack_float()
            i = data.unpack_float()
            poles.append(complex(r, i))
            r = data.unpack_float()
            i = data.unpack_float()
            zeros.append(complex(r, i))
        # first value describes number of poles/zeros
        npoles = int(poles[0].real) + 1
        nzeros = int(zeros[0].real) + 1
        ah_stats.station.poles = poles[1:npoles]
        ah_stats.station.zeros = zeros[1:nzeros]

        # event info
        ah_stats.event.latitude = data.unpack_float()
        ah_stats.event.longitude = data.unpack_float()
        ah_stats.event.depth = data.unpack_float()
        ot_year = data.unpack_int()
        ot_mon = data.unpack_int()
        ot_day = data.unpack_int()
        ot_hour = data.unpack_int()
        ot_min = data.unpack_int()
        ot_sec = data.unpack_float()
        try:
            ot = UTCDateTime(ot_year, ot_mon, ot_day, ot_hour, ot_min, ot_sec)
        except Exception:
            ot = None
        ah_stats.event.origin_time = ot
        ah_stats.event.comment = _unpack_string(data)

        # record info
        ah_stats.record.type = dtype = data.unpack_int()  # data type
        ah_stats.record.ndata = ndata = data.unpack_uint()  # number of samples
        ah_stats.record.delta = data.unpack_float()  # sampling interval
        ah_stats.record.max_amplitude = data.unpack_float()
        at_year = data.unpack_int()
        at_mon = data.unpack_int()
        at_day = data.unpack_int()
        at_hour = data.unpack_int()
        at_min = data.unpack_int()
        at_sec = data.unpack_float()
        at = UTCDateTime(at_year, at_mon, at_day, at_hour, at_min, at_sec)
        ah_stats.record.start_time = at
        ah_stats.record.abscissa_min = data.unpack_float()
        ah_stats.record.comment = _unpack_string(data)
        ah_stats.record.log = _unpack_string(data)

        # extras
        ah_stats.extras = data.unpack_array(data.unpack_float)

        # unpack data using dtype from record info
        if dtype == 1:
            # float
            temp = data.unpack_farray(ndata, data.unpack_float)
        elif dtype == 6:
            # double
            temp = data.unpack_farray(ndata, data.unpack_double)
        else:
            # e.g. 3 (vector), 2 (complex), 4 (tensor)
            msg = 'Unsupported AH v1 record type %d'
            raise NotImplementedError(msg % (dtype))
        tr = Trace(np.array(temp))
        tr.stats.ah = ah_stats
        tr.stats.delta = ah_stats.record.delta
        tr.stats.starttime = ah_stats.record.start_time
        tr.stats.station = ah_stats.station.code
        tr.stats.channel = ah_stats.station.channel
        return tr

    st = Stream()
    with open(filename, "rb") as fh:
        # read with XDR library
        data = xdrlib.Unpacker(fh.read())
        # loop as long we can read records
        while True:
            try:
                tr = _unpack_trace(data)
                st.append(tr)
            except EOFError:
                break
        return st


def _write_ah1(stream, filename):
    """
    Writes a Stream object to an AH v1 waveform file.

    :type stream:
    :param stream: The ObsPy Stream object to write.
    :type filename: str
    :param filename: open file, or file-like object

    """
    packer = xdrlib.Packer()

    for tr in stream:
        if hasattr(tr.stats, 'ah'):
            packer = _pack_trace_with_ah_dict(
                tr, packer, AH1_CODESIZE, AH1_CHANSIZE, AH1_STYPESIZE,
                AH1_COMSIZE, AH1_LOGSIZE)
        else:
            packer = _pack_trace_wout_ah_dict(
                tr, packer, AH1_CODESIZE, AH1_CHANSIZE, AH1_STYPESIZE,
                AH1_COMSIZE, AH1_LOGSIZE)

    with open(filename, 'wb') as fh:
        fh.write(packer.get_buffer())


def _pack_trace_with_ah_dict(tr, packer, codesize, chansize,
                             stypesize, comsize, logsize):

    # station info
    packer.pack_int(codesize)
    packer.pack_fstring(codesize, tr.stats.station.encode('utf-8'))
    packer.pack_int(chansize)
    packer.pack_fstring(chansize, tr.stats.channel.encode('utf-8'))

    packer.pack_int(stypesize)
    packer.pack_fstring(stypesize, tr.stats.ah.station.type.encode('utf-8'))
    packer.pack_float(tr.stats.ah.station.latitude)
    packer.pack_float(tr.stats.ah.station.longitude)
    packer.pack_float(tr.stats.ah.station.elevation)
    packer.pack_float(tr.stats.ah.station.gain)
    packer.pack_float(tr.stats.ah.station.normalization)

    poles = tr.stats.ah.station.poles
    zeros = tr.stats.ah.station.zeros

    # Poles and Zeros
    packer.pack_float(len(poles))
    packer.pack_float(0)
    packer.pack_float(len(zeros))
    packer.pack_float(0)

    for _i in range(1, 30):
        try:
            r, i = poles[_i].real, poles[_i].imag
        except IndexError:
            r, i = 0, 0
        packer.pack_float(r)
        packer.pack_float(i)

        try:
            r, i = zeros[_i].real, zeros[_i].imag
        except IndexError:
            r, i = 0, 0
        packer.pack_float(r)
        packer.pack_float(i)

    # event info
    packer.pack_float(tr.stats.ah.event.latitude)
    packer.pack_float(tr.stats.ah.event.longitude)
    packer.pack_float(tr.stats.ah.event.depth)
    if tr.stats.ah.event.origin_time is not None:
        packer.pack_int(tr.stats.ah.event.origin_time.year)
        packer.pack_int(tr.stats.ah.event.origin_time.month)
        packer.pack_int(tr.stats.ah.event.origin_time.day)
        packer.pack_int(tr.stats.ah.event.origin_time.hour)
        packer.pack_int(tr.stats.ah.event.origin_time.minute)
        packer.pack_float(tr.stats.ah.event.origin_time.second)
    else:
        packer.pack_int(0)
        packer.pack_int(0)
        packer.pack_int(0)
        packer.pack_int(0)
        packer.pack_int(0)
        packer.pack_float(0)

    packer.pack_int(comsize)
    packer.pack_fstring(comsize, tr.stats.ah.event.comment.encode('utf-8'))

    # record info
    dtype = tr.stats.ah.record.type
    packer.pack_int(dtype)
    ndata = tr.stats.npts
    packer.pack_uint(ndata)
    packer.pack_float(tr.stats.ah.record.delta)
    packer.pack_float(tr.stats.ah.record.max_amplitude)
    packer.pack_int(tr.stats.ah.record.start_time.year)
    packer.pack_int(tr.stats.ah.record.start_time.month)
    packer.pack_int(tr.stats.ah.record.start_time.day)
    packer.pack_int(tr.stats.ah.record.start_time.hour)
    packer.pack_int(tr.stats.ah.record.start_time.minute)
    packer.pack_float(tr.stats.ah.record.start_time.second)
    packer.pack_float(tr.stats.ah.record.abscissa_min)
    packer.pack_int(comsize)
    packer.pack_fstring(comsize, tr.stats.ah.record.comment.encode('utf-8'))
    packer.pack_int(logsize)
    packer.pack_fstring(logsize, tr.stats.ah.record.log.encode('utf-8'))

    # # extras
    packer.pack_array(tr.stats.ah.extras, packer.pack_float)

    # pack data using dtype from record info
    if dtype == 1:
        # float
        packer.pack_farray(ndata, tr.data, packer.pack_float)
    elif dtype == 6:
        # double
        packer.pack_farray(ndata, tr.data, packer.pack_double)
    else:
        # e.g. 3 (vector), 2 (complex), 4 (tensor)
        msg = 'Unsupported AH v1 record type %d'
        raise NotImplementedError(msg % (dtype))

    return packer


def _pack_trace_wout_ah_dict(tr, packer, codesize, chansize,
                             stypesize, comsize, logsize):
    """
    Entry are packed in the same order as shown in
    _pack_trace_with_ah_dict .The missing information
    is replaced with zeros
    station info
    """
    packer.pack_int(codesize)
    packer.pack_fstring(codesize, tr.stats.station.encode('utf-8'))
    packer.pack_int(chansize)
    packer.pack_fstring(chansize, tr.stats.channel.encode('utf-8'))
    packer.pack_int(stypesize)
    packer.pack_fstring(stypesize, 'null'.encode('utf-8'))
    # There is no information about latitude, longitude, elevation,
    # gain and normalization in the basic stream object,  are set to 0
    packer.pack_float(0)
    packer.pack_float(0)
    packer.pack_float(0)
    packer.pack_float(0)
    packer.pack_float(0)

    # Poles and Zeros are not provided by stream object, are set to 0
    for _i in range(0, 30):
        packer.pack_float(0)
        packer.pack_float(0)
        packer.pack_float(0)
        packer.pack_float(0)

    # event info
    packer.pack_float(0)
    packer.pack_float(0)
    packer.pack_float(0)
    packer.pack_int(0)
    packer.pack_int(0)
    packer.pack_int(0)
    packer.pack_int(0)
    packer.pack_int(0)
    packer.pack_float(0)

    packer.pack_int(comsize)
    packer.pack_fstring(comsize, 'null'.encode('utf-8'))

    # record info
    dtype = type(tr.data[0])
    if '32' in str(dtype):
        dtype = 1
    elif '64' in str(dtype):
        dtype = 6

    packer.pack_int(dtype)
    ndata = tr.stats.npts
    packer.pack_uint(ndata)
    packer.pack_float(tr.stats.delta)
    packer.pack_float(max(tr.data))
    packer.pack_int(tr.stats.starttime.year)
    packer.pack_int(tr.stats.starttime.month)
    packer.pack_int(tr.stats.starttime.day)
    packer.pack_int(tr.stats.starttime.hour)
    packer.pack_int(tr.stats.starttime.minute)

    sec = tr.stats.starttime.second
    msec = tr.stats.starttime.microsecond
    starttime_second = float(str(sec) + '.' + str(msec))
    packer.pack_float(starttime_second)

    packer.pack_float(0)
    packer.pack_int(comsize)
    packer.pack_fstring(comsize, 'null'.encode('utf-8'))
    packer.pack_int(logsize)
    packer.pack_fstring(logsize, 'null'.encode('utf-8'))

    # # extras
    packer.pack_array(np.zeros(21).tolist(), packer.pack_float)

    # pack data using dtype from record info
    if dtype == 1:
        # float
        packer.pack_farray(ndata, tr.data, packer.pack_float)
    elif dtype == 6:
        # double
        packer.pack_farray(ndata, tr.data, packer.pack_double)
    else:
        # e.g. 3 (vector), 2 (complex), 4 (tensor)
        msg = 'Unsupported AH v1 record type %d'
        raise NotImplementedError(msg % (dtype))

    return packer


def _read_ah2(filename):
    """
    Reads an AH v2 waveform file and returns a Stream object.

    :type filename: str
    :param filename: AH v2 file to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream with Traces specified by given file.
    """

    def _unpack_trace(data):
        ah_stats = AttribDict({
            'version': '2.0',
            'event': AttribDict(),
            'station': AttribDict(),
            'record': AttribDict(),
            'extras': []
        })

        # station info
        data.unpack_int()  # undocumented extra int?
        ah_stats.station.code = _unpack_string(data)
        data.unpack_int()  # here too?
        ah_stats.station.channel = _unpack_string(data)
        data.unpack_int()  # and again?
        ah_stats.station.type = _unpack_string(data)
        ah_stats.station.recorder = _unpack_string(data)
        ah_stats.station.sensor = _unpack_string(data)
        ah_stats.station.azimuth = data.unpack_float()  # degrees E from N
        ah_stats.station.dip = data.unpack_float()  # up = -90, down = +90
        ah_stats.station.latitude = data.unpack_double()
        ah_stats.station.longitude = data.unpack_double()
        ah_stats.station.elevation = data.unpack_float()
        ah_stats.station.gain = data.unpack_float()
        ah_stats.station.normalization = data.unpack_float()  # A0

        npoles = data.unpack_int()
        ah_stats.station.poles = []
        for _i in range(npoles):
            r = data.unpack_float()
            i = data.unpack_float()
            ah_stats.station.poles.append(complex(r, i))

        nzeros = data.unpack_int()
        ah_stats.station.zeros = []
        for _i in range(nzeros):
            r = data.unpack_float()
            i = data.unpack_float()
            ah_stats.station.zeros.append(complex(r, i))
        ah_stats.station.comment = _unpack_string(data)

        # event info
        ah_stats.event.latitude = data.unpack_double()
        ah_stats.event.longitude = data.unpack_double()
        ah_stats.event.depth = data.unpack_float()
        ot_year = data.unpack_int()
        ot_mon = data.unpack_int()
        ot_day = data.unpack_int()
        ot_hour = data.unpack_int()
        ot_min = data.unpack_int()
        ot_sec = data.unpack_float()
        try:
            ot = UTCDateTime(ot_year, ot_mon, ot_day, ot_hour, ot_min, ot_sec)
        except Exception:
            ot = None
        ah_stats.event.origin_time = ot
        data.unpack_int()  # and again?
        ah_stats.event.comment = _unpack_string(data)

        # record info
        ah_stats.record.type = dtype = data.unpack_int()  # data type
        ah_stats.record.ndata = ndata = data.unpack_uint()  # number of samples
        ah_stats.record.delta = data.unpack_float()  # sampling interval
        ah_stats.record.max_amplitude = data.unpack_float()
        at_year = data.unpack_int()
        at_mon = data.unpack_int()
        at_day = data.unpack_int()
        at_hour = data.unpack_int()
        at_min = data.unpack_int()
        at_sec = data.unpack_float()
        at = UTCDateTime(at_year, at_mon, at_day, at_hour, at_min, at_sec)
        ah_stats.record.start_time = at
        ah_stats.record.units = _unpack_string(data)
        ah_stats.record.inunits = _unpack_string(data)
        ah_stats.record.outunits = _unpack_string(data)
        data.unpack_int()  # and again?
        ah_stats.record.comment = _unpack_string(data)
        data.unpack_int()  # and again?
        ah_stats.record.log = _unpack_string(data)

        # user attributes
        nusrattr = data.unpack_int()
        ah_stats.usrattr = {}
        for _i in range(nusrattr):
            key = _unpack_string(data)
            value = _unpack_string(data)
            ah_stats.usrattr[key] = value

        # unpack data using dtype from record info
        if dtype == 1:
            # float
            temp = data.unpack_farray(ndata, data.unpack_float)
        elif dtype == 6:
            # double
            temp = data.unpack_farray(ndata, data.unpack_double)
        else:
            # e.g. 3 (vector), 2 (complex), 4 (tensor)
            msg = 'Unsupported AH v2 record type %d'
            raise NotImplementedError(msg % (dtype))

        tr = Trace(np.array(temp))
        tr.stats.ah = ah_stats
        tr.stats.delta = ah_stats.record.delta
        tr.stats.starttime = ah_stats.record.start_time
        tr.stats.station = ah_stats.station.code
        tr.stats.channel = ah_stats.station.channel
        return tr

    st = Stream()
    with open(filename, "rb") as fh:
        # loop as long we can read records
        while True:
            try:
                # read first 8 bytes with XDR library
                data = xdrlib.Unpacker(fh.read(8))
                # check magic version number
                magic = data.unpack_int()
            except EOFError:
                break
            if magic != 1100:
                raise Exception('Not a AH v2 file')
            try:
                # get record length
                length = data.unpack_uint()
                # read rest of record into XDR unpacker
                data = xdrlib.Unpacker(fh.read(length))
                tr = _unpack_trace(data)
                st.append(tr)
            except EOFError:
                break
        return st
